#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "go_board.h"
#include "go_string.h"


/**
 * This function initilaizes go_board structure.
 *
 * \param board  Structure of go_board
 * \param size   Size of the board
 *
 * \retval 0  Success
 */
int go_board_initialize(struct go_board* board, int size)
{
    board->board_size = size;
    board->num_captured_white = 0;
    board->num_captured_black = 0;
    board->ko_x = STONE_COORD_INVALID;
    board->ko_y = STONE_COORD_INVALID;

    for (coord_t y = 0; y < size; y++) {
        for (coord_t x = 0; x < size; x++) {
            board->board_array[y][x] = 0;
        }
    }

    struct go_string_list* lst =
        (struct go_string_list*)malloc(sizeof(struct go_string_list));
    if (lst == NULL) {
        return -1;
    }
    board->string_list = lst;
    go_string_list_initialize(board->string_list, size);

    return 0;
}


/**
 * This function terminates go_board structure
 *
 * \retval 0  Success
 */
int go_board_terminate(struct go_board* board)
{
    go_string_list_terminate(board->string_list);
    if (board->string_list != NULL) {
        free(board->string_list);
        board->string_list = NULL;
    }

    return 0;
}


/**
 * This function copies the board status of \a src to \a dst.
 *
 * \param \a dst  Board which is the destination of copy
 * \param \a src  Board which is the source of copy
 *
 * \retval 0  Success
 */
int go_board_copy_board(struct go_board* dst, struct go_board* src)
{
    assert(dst->board_size == src->board_size);

    const int BOARD_SIZE = src->board_size;

    for (coord_t y = 0; y < BOARD_SIZE; y++) {
        for (coord_t x = 0; x < BOARD_SIZE; x++) {
            dst->board_array[y][x] = src->board_array[y][x];
        }
    }
    dst->num_captured_white = src->num_captured_white;
    dst->num_captured_black = src->num_captured_black;
    dst->ko_x = src->ko_x;
    dst->ko_y = src->ko_y;

    /*
     * Copy string_list
     */
    go_string_list_copy_list(dst->string_list, src->string_list);

    return 0;
}


/**
 * This function gets the status of ( \a x , \a y ).
 *
 * \param x  Coordinate of x-axis
 * \param y  Coordinate of y-axis
 *
 * \retval STONE_COLOR_EMPTY  No stone
 * \retval STONE_COLOR_BLACK  Black stone
 * \retval STONE_COLOR_WHITE  White stone
 */
stone_t go_board_get_stone(struct go_board* board, coord_t x, coord_t y)
{
    return board->board_array[y][x];
}


/**
 * This function sets the stone to ( \a x , \a y ).
 *
 * \param color  Color of stone
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval 0   Success
 * \retval -1  Invalid coordinate
 */
int go_board_set_stone(struct go_board* board, stone_t color, coord_t x,
		       coord_t y)
{
    board->board_array[y][x] = color;
    return 0;
}


/**
 * This function removes the stone from ( \a x , \a y ).
 *
 * \param x  Coordinate of x-axis
 * \param y  Coordinate of y-axis
 *
 * \retval 0   Success
 * \retval -1  Invalid coordinate
 */
int go_board_remove_stone(struct go_board* board, coord_t x, coord_t y)
{
    board->board_array[y][x] = STONE_COLOR_EMPTY;
    return 0;
}


/**
 * This function returns the size of the board.
 *
 * \retval positive_number  Board size
 */
int go_board_get_board_size(struct go_board* board)
{
    return board->board_size;
}


/**
 * Increment the number of captured white stones.
 *
 * \retval 0  Success
 */
int go_board_increment_captured_white(struct go_board* board)
{
    board->num_captured_white++;
    return 0;
}


/**
 * Increment the number of captured black stones.
 *
 * \retval 0  Success
 */
int go_board_increment_captured_black(struct go_board* board)
{
    board->num_captured_black++;
    return 0;
}


/**
 * Get the number of captured white stones.
 */
int go_board_get_captured_white(struct go_board* board)
{
    return board->num_captured_white;
}


/**
 * Get the number of captured black stones.
 */
int go_board_get_captured_black(struct go_board* board)
{
    return board->num_captured_black;
}


/**
 * Get the number of black stones on the board.
 */
int go_board_get_black(struct go_board* board)
{
    const int board_size = board->board_size;

    int num_black = 0;
    for (coord_t y = 0; y < board_size; y++) {
        for (coord_t x = 0; x < board_size; x++) {
            if (board->board_array[y][x] == STONE_COLOR_BLACK) {
                num_black++;
            }
        }
    }
    return num_black;
}


/**
 * Get the number of white stones on the board.
 */
int go_board_get_white(struct go_board* board)
{
    const int board_size = board->board_size;

    int num_white = 0;
    for (coord_t y = 0; y < board_size; y++) {
        for (coord_t x = 0; x < board_size; x++) {
            if (board->board_array[y][x] == STONE_COLOR_WHITE) {
                num_white++;
            }
        }
    }
    return num_white;
}


/**
 * Set the ko position.
 *
 * \param x  Coordinate of x-axis
 * \param y  Coordinate of y-axis
 *
 * \retval 0  Success
 */
int go_board_set_ko_position(struct go_board* board, coord_t x, coord_t y)
{
    board->ko_x = x;
    board->ko_y = y;
    return 0;
}


/**
 * Get the ko position.
 *
 * \param x  (OUT) Coordinate of x-axis
 * \param y  (OUT) Coordinate of y-axis
 *
 * \retval 0  Success
 */
int go_board_get_ko_position(struct go_board* board, coord_t* x, coord_t* y)
{
    *x = board->ko_x;
    *y = board->ko_y;
    return 0;
}


/**
 * This function dumps the status of the board. (For debugging)
 *
 * \retval 0  Success
 */
int go_board_display(struct go_board* board)
{
    const int board_size = board->board_size;
    const char stone_char[] = { '+', '#', 'O' };

    {
        printf("    ");
        for (coord_t x = 0; x < board_size; x++) {
            if (x < 8) {
                printf("%2c", x + 'A'); /* A, B, ... H */
            } else {
                printf("%2c", x + 'A' + 1); /* J, K, ... */
            }
        }
        printf("\n");
    }

    for (coord_t y = 0; y < board_size; y++) {
        printf("[%2d] ", board_size - y);
        for (coord_t x = 0; x < board_size; x++) {
	    const stone_t stone = go_board_get_stone(board, x, y);
            printf("%c ", stone_char[stone]);
        }
        printf("\n");
    }
    printf("\n");

    printf("Captured stones:\n");
    printf("- White stones captured by Black stone  %d [stones]\n",
           go_board_get_captured_white(board));
    printf("- Black stones captured by White stone  %d [stones]\n",
           go_board_get_captured_black(board));
    printf("\n");

    return 0;
}
