#include "go_board.h"
#include "go_engine.h"
#include "go_engine_internal.h"
#include "go_engine_mc.h"
#include "go_string.h"

#include "sgf_loader.h"
#include "utility.h"

#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define MAX_EXEC_TIME_DEFAULT   (850*1000) /* 850 msec */

#define THRESHOLD_WIN_RATE      0.50
#define THRESHOLD_WIN_RATE_HIGH 0.60


/*
 * Exported functions for engine codes
 */
extern stone_t my_color_get(void);
extern stone_t opponent_color_get(void);
extern stone_t reverse_color_get(stone_t);
extern int go_engine_get_board_size(void);
extern float go_engine_get_komi(void);
extern int64_t go_engine_set_start_time(void); // XXX: Only for tests
extern int64_t go_engine_get_start_time(void);
extern int64_t go_engine_get_max_exec_time(void);
extern int go_engine_get_prev_pos(coord_t*, coord_t*);
struct go_board* go_engine_get_main_board(void);
struct go_board* go_engine_get_work_board(void);
struct go_board* go_engine_get_work_board_rear(void); // XXX: For MC

/*
 * Imported functions from go_engine_subr.c
 */
extern bool is_empty_position(struct go_board*, coord_t, coord_t);
extern bool is_all_same_color(struct go_board*, stone_t);
extern bool is_capture_position(struct go_board*, stone_t, coord_t, coord_t);
extern bool is_surrounded_position_single(struct go_board*, stone_t, coord_t,
                                          coord_t);
/*
 * Internal functions
 */
static bool engine_is_configurable(void);


/*
 * Engine structure
 */
static struct go_engine g_engine;
static bool g_is_engine_initialized = false;


/**
 * This function initializes the go_engine instance.
 *
 * \param size   Board size
 * \param color  Color of my stone (STONE_COLOR_{BLACK,WHITE})
 * \param komi   Value of komi
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_engine_initialize(int size, stone_t color, float komi)
{
    if (g_is_engine_initialized) {
        return -1;
    }

    g_engine.board_size = size;
    g_engine.my_stone_color = color;
    g_engine.move_count = 0;
    g_engine.komi = komi;
    g_engine.max_exec_time = MAX_EXEC_TIME_DEFAULT;
    g_engine.prev_pos.x = STONE_COORD_INVALID;
    g_engine.prev_pos.y = STONE_COORD_INVALID;

    go_board_initialize(&(g_engine.gboard_main), size);

    for (int i = 0; i < NUM_BOARD_COPY; i++) {
        go_board_initialize(&(g_engine.gboard_copy[i]), size);
    }

    g_is_engine_initialized = true;

    //srand((unsigned)time(NULL));

    return 0;
}


/**
 * This function terminates the go_engine instance.
 *
 * \retval 0  Success
 */
int go_engine_terminate(void)
{
    if (!g_is_engine_initialized) {
        return -1;
    } else {
        go_board_terminate(&(g_engine.gboard_main));

        for (int i = 0; i < NUM_BOARD_COPY; i++) {
            go_board_terminate(&(g_engine.gboard_copy[i]));
        }

        g_is_engine_initialized = false;

        return 0;
    }
}


/**
 * This function re-initializes the go_engine instance.
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_engine_reinitialize(void)
{
    if (!g_is_engine_initialized) {
        return -1;
    }

    /*
     * Initialize the move count, the previous move and the boards only.
     * Other parameters have to be preserved.
     */
    g_engine.move_count = 0;
    g_engine.prev_pos.x = STONE_COORD_INVALID;
    g_engine.prev_pos.y = STONE_COORD_INVALID;

    const int BOARD_SIZE = g_engine.board_size;

    go_board_initialize(&(g_engine.gboard_main), BOARD_SIZE);
    for (int i = 0; i < NUM_BOARD_COPY; i++) {
        go_board_initialize(&(g_engine.gboard_copy[i]), BOARD_SIZE);
    }

    return 0;
}


/**
 * This function sets the information on the opponent's move to the engine.
 *
 * \param flags  Flags for the move (MOVE_FLAGS_{PASS, RESIGN, NORMAL})
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval 1   Pass
 * \retval 0   Normal move
 * \retval -1  Resign
 *
 * \note This function uses string functions indirectly.
 */
int go_engine_set_opponent_move(move_t flags, coord_t x, coord_t y)
{
    if (g_engine.move_count == 0) {
        g_engine.my_stone_color = STONE_COLOR_WHITE;
    }

    g_engine.move_count++;

    if (flags == MOVE_FLAGS_PASS) {
        g_engine.prev_pos.x = STONE_COORD_INVALID;
        g_engine.prev_pos.y = STONE_COORD_INVALID;
        return 1;
    } else if (flags == MOVE_FLAGS_RESIGN) {
        return -1;
    } else {
        stone_place(&(g_engine.gboard_main), opponent_color_get(), x, y);
        g_engine.prev_pos.x = x;
        g_engine.prev_pos.y = y;

        return 0;
    }
}


/**
 * This function gets the information on the next move.
 *
 * It evaluates all the possible moves and find the best one.
 * The best move is returned by this function.
 *
 * \param flags  (OUT) Flags for move (MOVE_INFO_FLAGS_{NORMAL,PASS})
 * \param rx     (OUT) Coordinate of x-axis generated by this function
 * \parma ry     (OUT) Coordinate of y-axis generated by this function
 *
 * \retval 0  Success (normal move)
 * \retval 1  Success (pass)
 *
 * \note This function uses string functions indirectly.
 */
int go_engine_get_next_move(move_t* flags, coord_t* rx, coord_t* ry)
{
    int rv;
    bool is_valid_candidate = false;

    coord_t best_x, best_y;
    double score;

    const stone_t my_color = my_color_get();
    g_engine.start_time = current_time_get();

    const int count = g_engine.move_count;
    rv = go_engine_search_mc(my_color, count, &best_x, &best_y, &score);
    if (rv >= 0) {
        is_valid_candidate = true;
    }

    g_engine.move_count++;

    /*
     * There are no valid candidates.
     */
    if (!is_valid_candidate) {
        *flags = MOVE_FLAGS_PASS;
        *rx = STONE_COORD_INVALID;
        *ry = STONE_COORD_INVALID;
        return 1;
    }

    /*
     * If the score of the best candidate is better than the threshold,
     * place the stone. Otherwise just pass.
     */
    double threshold;
    if (rv == 0) {
        threshold = THRESHOLD_WIN_RATE;
    } else {
        assert(rv == 1);
        threshold = THRESHOLD_WIN_RATE_HIGH;
    }
    if (score >= threshold) {
        *flags = MOVE_FLAGS_NORMAL;
        *rx = best_x;
        *ry = best_y;
        stone_place(&(g_engine.gboard_main), my_color, best_x, best_y);
        return 0;
    } else {
        *flags = MOVE_FLAGS_PASS;
        *rx = STONE_COORD_INVALID;
        *ry = STONE_COORD_INVALID;
        return 1;
    }
}


/**
 * This function gets the current score.
 *
 * \param score  (OUT) Current score
 *
 * \retval 0   Success
 */
int go_engine_get_current_score(score_t* score)
{
    score->black_score = 0;
    score->white_score = 0;

    return 0;
}


/**
 * This function sets the board size.
 *
 * \param size  Board size
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_engine_set_boardsize(int size)
{
    if (!engine_is_configurable()) {
        return -1;
    } else {
        g_engine.board_size = size;
        return 0;
    }
}


/**
 * This function sets the value of komi.
 *
 * \param komi  Value of komi
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_engine_set_komi(float komi)
{
    if (!engine_is_configurable()) {
        return -1;
    } else {
        g_engine.komi = komi;
        return 0;
    }
}


/**
 * This function sets time-settings.
 *
 * \param mtime  Main time in sec
 * \param atime  Additional time (= byo-yomi time) in usec
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_engine_set_time_settings(int mtime, int atime)
{
    (void)mtime; /* XXX: Ignore main time */

    if (!engine_is_configurable()) {
        return -1;
    } else {
        g_engine.max_exec_time = atime;
        return 0;
    }
}


/**
 * This function loads the specified SGF file, and updates the board status.
 *
 * If MOVE_COUNT_INVALID is specified to \a move_num, the board status is
 * updated to the latest status described by the SGF file.
 *
 * \param filename  SGF filename
 * \param move_num  Number of moves
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_engine_load_sgf_file(const char* filename, int move_num)
{
    int rv;
    int board_size;
    float komi;
    const int BLIST_NUM = MAX_BOARD_SIZE * MAX_BOARD_SIZE;
    const int WLIST_NUM = MAX_BOARD_SIZE * MAX_BOARD_SIZE;
    coord2d_t blist[BLIST_NUM];
    coord2d_t wlist[WLIST_NUM];

    /*
     * \a move_num should be MOVE_COUNT_INVALID at this point.
     */
    if (move_num != MOVE_COUNT_INVALID) {
        return -1;
    }

    if (!engine_is_configurable()) {
        return -1;
    }

    int blist_num = BLIST_NUM;
    int wlist_num = WLIST_NUM;
    rv = sgf_file_load_simple(filename, &board_size, &komi, blist,
                              &blist_num, wlist, &wlist_num);
    if (rv != 0) {
        return -1;
    }

    /*
     * Update engine's status
     */
    {
        g_engine.board_size = board_size;
        g_engine.komi = komi;

        go_engine_reinitialize();

        for (int i = 0; i < blist_num; i++) {
            go_board_set_stone(&(g_engine.gboard_main), STONE_COLOR_BLACK,
                               blist[i].x, blist[i].y);
        }
        for (int i = 0; i < wlist_num; i++) {
            go_board_set_stone(&(g_engine.gboard_main), STONE_COLOR_WHITE,
                               wlist[i].x, wlist[i].y);
        }
    }

    return 0;
}


/***
 ***
 *** Internal functions which are exported to engine codes.
 ***
 ***/

/** \internal
 * This function returns the color of my stone.
 *
 * \retval STONE_COLOR_BLACK  Black stone
 * \retval STONE_COLOR_WHITE  White stone
 */
stone_t my_color_get(void)
{
    return g_engine.my_stone_color;
}


/** \internal
 * This function returns the color of the opponent's stone.
 *
 * \retval STONE_COLOR_BLACK  Black stone
 * \retval STONE_COLOR_WHITE  White stone
 */
stone_t opponent_color_get(void)
{
    if (g_engine.my_stone_color == STONE_COLOR_BLACK) {
	return STONE_COLOR_WHITE;
    } else {
	return STONE_COLOR_BLACK;
    }
}


/** \internal
 * This function returns the reverse color of the given stone.
 *
 * \param color  Color of stone (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 *
 * \retval STONE_COLOR_BLACK  Black stone (White stone is specified.)
 * \retval STONE_COLOR_WHITE  White stone (Black stone is specified.)
 */
stone_t reverse_color_get(stone_t color)
{
    if (color == STONE_COLOR_BLACK) {
	return STONE_COLOR_WHITE;
    } else {
	return STONE_COLOR_BLACK;
    }
}


/** \internal
 * This function returns the board size.
 *
 * \retval board_size  Board size
 */
int go_engine_get_board_size(void)
{
    return g_engine.board_size;
}


/** \internal
 * This function returns the komi.
 *
 * \retval komi  Komi
 */
float go_engine_get_komi(void)
{
    return g_engine.komi;
}


/** \internal
 * This function sets the start time of this turn. (only for test)
 *
 * \retval start_time  Start time of this turn in usec
 */
int64_t go_engine_set_start_time(void)
{
    g_engine.start_time = current_time_get();
    return g_engine.start_time;
}


/** \internal
 * This function returns the start time of this turn.
 *
 * \retval start_time  Start time of this turn in usec
 */
int64_t go_engine_get_start_time(void)
{
    return g_engine.start_time;
}


/** \internal
 * This function returns the maximum execution time.
 *
 * \retval max_exec_time  Maximum execution time in usec
 */
int64_t go_engine_get_max_exec_time(void)
{
    return g_engine.max_exec_time;
}


/** \internal
 * This function returns the previous position of the opponent.
 *
 * \param x  (OUT) Previous position (X-coordinate)
 * \param y  (OUT) Previous position (Y-coordinate)
 *
 * \retval 0  Success
 */
int go_engine_get_prev_pos(coord_t* x, coord_t* y)
{
    *x = g_engine.prev_pos.x;
    *y = g_engine.prev_pos.y;
    return 0;
}


/** \internal
 * This function returns the pointer to the main board.
 *
 * \retval main_board  Pointer to the main board
 */
struct go_board* go_engine_get_main_board(void)
{
    return &(g_engine.gboard_main);
}


/** \internal
 * This function returns the pointer to the working board.
 *
 * \retval work_board  Pointer to the working board.
 */
struct go_board* go_engine_get_work_board(void)
{
    return &(g_engine.gboard_copy[0]);
}


/** \internal
 * This function returns the pointer to the working board (rear).
 *
 * \retval work_board  Pointer to the working board (rear).
 */
struct go_board* go_engine_get_work_board_rear(void)
{
    return &(g_engine.gboard_copy[1]);
}


/***
 ***
 *** Internal functions for this file
 ***
 ***/

/** \internal
 * This function determines whether the configuration of the engine
 * can be changed or not.
 *
 * The configuration can be changed only if the engine is initialized
 * and no moves are performed. (That is, the game is not started yet.)
 *
 * \retval true   The configuration can be changed.
 * \retval false  The configuration cannot be changed.
 */
bool engine_is_configurable(void)
{
    if ((g_is_engine_initialized) && (g_engine.move_count == 0)) {
        return true;
    } else {
        return false;
    }
}
