#include "go_engine_mc.h"
#include "go_engine_internal.h"

#include "uct_tree.h"
#include "utility.h"

#include <assert.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*#define ENABLE_PLAYOUT_PROFILE*/
/*#define DIAGNOSTIC*/
/*#define MEMLEAK_WORKAROUND*/
/*#define UCT_DEBUG*/

#define MAX_PLAYOUT_MOVE_NUM    100
#define UCT_CREATE_NODE_LIMIT   20


/*
 * Tree structure for UCT
 * XXX: Should move g_uct_tree to the engine structure.
 */
static struct uct_tree g_uct_tree;


/*
 * Temporal buffers
 */
#define NUM_MOVE_ARRAY          (MAX_BOARD_SIZE * MAX_BOARD_SIZE)
#define NUM_NODE_ARRAY          1024
static coord2d_t move_array[NUM_MOVE_ARRAY];
static coord2d_t move_array_xxx[NUM_MOVE_ARRAY];
static struct uct_node* node_array[NUM_NODE_ARRAY];


#ifdef ENABLE_PLAYOUT_PROFILE
/*
 * Playout profile (for debugging)
 */
static int num_playout = 0;
#endif /* ENABLE_PLAYOUT_PROFILE */


/*
 * Imported from go_engine.c
 */
extern int go_engine_get_board_size(void);
extern int64_t go_engine_get_start_time(void);
extern int64_t go_engine_get_max_exec_time(void);
extern int go_engine_get_prev_pos(coord_t*, coord_t*);
extern float go_engine_get_komi(void);
extern struct go_board* go_engine_get_main_board(void);
extern struct go_board* go_engine_get_work_board(void);
extern struct go_board* go_engine_get_work_board_rear(void);
extern stone_t my_color_get(void);
extern stone_t reverse_color_get(stone_t);

/*
 * Internal functions (but exported for tests)
 */
int playOneSequence(struct uct_tree*, int64_t);
struct uct_node* descendByUCB1(struct uct_node*);
int updateValue(struct uct_node**, int, double);
double getValueByMC(struct uct_node*, struct go_board*, stone_t);

/*
 * Internal functions
 */
static int mc_playout(struct go_board*, stone_t, coord_t, coord_t, double*);
static int mc_get_candidate_move(struct go_board*, stone_t, coord_t*,
                                 coord_t*);
static float mc_get_match_result(struct go_board*, stone_t);
static int mc_get_random_number(int, int);


/**
 * This function searches the best move by UCT method.
 *
 * If there is at least a valid move, this function returns 0.
 * The position of the best move is stored to \a x and \a y.
 * The score (winning rate) is also stored to \a score.
 *
 * Otherwise it returns a negative value.
 *
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param count  Current move count (>= 0)
 * \param x      (OUT) Position of the move (X coordinate)
 * \param y      (OUT) Position of the move (Y coordinate)
 * \param score  (OUT) Score of the move
 *
 * \retval 0   Success
 * \retval 1   Success (The best move is not well simulated. Maybe bad.)
 * \retval -1  No valid moves are found.
 * \retval -2  Should pass because the best move seems to be a eye.
 */
int go_engine_search_mc(stone_t color, int count, coord_t* x, coord_t* y, double* score)
{
    int rv;

    const int64_t start_time = go_engine_get_start_time();
    const int64_t MAX_EXEC_TIME = go_engine_get_max_exec_time();
    const int64_t end_time = start_time + MAX_EXEC_TIME;
    struct go_board* orig_board = go_engine_get_main_board();
    struct go_board* work_board = go_engine_get_work_board();


#ifdef ENABLE_PLAYOUT_PROFILE
    /*
     * Initialize playout count
     */
    num_playout = 0;
    const int64_t t1 = current_time_get();
#endif /* ENABLE_PLAYOUT_PROFILE */


    /*
     * Check whether there is a valid move or not.
     */
    int num_move;
    rv = valid_move_get(orig_board, color, move_array_xxx, NUM_MOVE_ARRAY,
                        &num_move);
    if (rv != 0) {
        return -1;
    }


    /*
     * Select the candidates. (Only for 9x9 board)
     */
    {
        const int EARLY_STAGE_COUNT = 40;
        const int BOARD_SIZE = go_engine_get_board_size();

        /*
         * Place a position next to the center ("tengen").
         */
        if ((count == 0) && (color == STONE_COLOR_BLACK)) {
            const int TAKAMOKU_POS_X = 4;
            const int TAKAMOKU_POS_Y = 3;
            *x = TAKAMOKU_POS_X;
            *y = TAKAMOKU_POS_Y;
            *score = 1.0;
            return 0;
        }

        /*
         * Place a 3x3 position which is the widest for the white.
         */
        if ((count == 1) && (color == STONE_COLOR_WHITE)) {
            coord_t prev_x, prev_y;
            go_engine_get_prev_pos(&prev_x, &prev_y);

            /* Pass */
            if ((prev_x == STONE_COORD_INVALID)
                && (prev_y == STONE_COORD_INVALID)) {
                const int TAKAMOKU_POS_X = 4;
                const int TAKAMOKU_POS_Y = 5;
                *x = TAKAMOKU_POS_X;
                *y = TAKAMOKU_POS_Y;
                *score = 1.0;
                return 0;
            }
            /* Upper-left */
            if ((0 <= prev_x) && (prev_x <= BOARD_SIZE/2) && (0 <= prev_y)
                && (prev_y <= BOARD_SIZE/2)) {
                const int LR_3X3_POS_X = 6;
                const int LR_3X3_POS_Y = 6;
                *x = LR_3X3_POS_X;
                *y = LR_3X3_POS_Y;
                *score = 1.0;
                return 0;
            }
            /* Lower-left */
            if ((0 <= prev_x) && (prev_x <= BOARD_SIZE/2)
                && (BOARD_SIZE/2 < prev_y) && (prev_y < BOARD_SIZE)) {
                const int UR_3X3_POS_X = 6;
                const int UR_3X3_POS_Y = 2;
                *x = UR_3X3_POS_X;
                *y = UR_3X3_POS_Y;
                *score = 1.0;
                return 0;
            }
            /* Upper-right */
            if ((BOARD_SIZE/2 < prev_x) && (prev_x < BOARD_SIZE)
                && (0 <= prev_y) && (prev_y <= BOARD_SIZE/2)) {
                const int LL_3X3_POS_X = 2;
                const int LL_3X3_POS_Y = 6;
                *x = LL_3X3_POS_X;
                *y = LL_3X3_POS_Y;
                *score = 1.0;
                return 0;
            }
            /* Lower-right */
            if ((BOARD_SIZE/2 < prev_x) && (prev_x < BOARD_SIZE)
                && (BOARD_SIZE/2 < prev_y) && (prev_y < BOARD_SIZE)) {
                const int UL_3X3_POS_X = 2;
                const int UL_3X3_POS_Y = 2;
                *x = UL_3X3_POS_X;
                *y = UL_3X3_POS_Y;
                *score = 1.0;
                return 0;
            }
        }

        /*
         * Rule out one-line positions during the early stage.
         */
        int num_removed = 0;
        if (count < EARLY_STAGE_COUNT) {
            for (int i = 0; i < num_move; i++) {
                if ((move_array_xxx[i].x == 0)
                    || (move_array_xxx[i].y == 0)
                    || (move_array_xxx[i].x == (BOARD_SIZE-1))
                    || (move_array_xxx[i].y == (BOARD_SIZE-1))) {
                    move_array_xxx[i].x = STONE_COORD_INVALID;
                    move_array_xxx[i].y = STONE_COORD_INVALID;
                    num_removed++;
                }
            }
        }

        /*
         * Check stone connections
         */
#ifdef notyet
#endif /* notyet */

        /*
         * Sort move_array_xxx[].
         */
        for (int i = 0; i < num_move; i++) {
            for (int j = i+1; j < num_move; j++) {
                const int mx1 = move_array_xxx[i].x;
                const int mx2 = move_array_xxx[j].x;
                if (mx1 < mx2) {
                    const coord2d_t tmp = move_array_xxx[j];
                    move_array_xxx[j] = move_array_xxx[i];
                    move_array_xxx[i] = tmp;
                }
            }
        }
        num_move -= num_removed;
        assert(num_move >= 0);
    }


    /*
     * Initialize UCT tree.
     */
    rv = uct_tree_initialize(&g_uct_tree);
    assert(rv == 0);


    /*
     * Add all the valid moves to children of the root.
     */
    struct uct_node* root = g_uct_tree.root;
    for (int j = 0; j < num_move; j++) {
        struct uct_node* new_node;
        rv = uct_tree_create_node(&g_uct_tree, &new_node);
        assert(rv == 0);
        rv = uct_tree_add_child(&g_uct_tree, root, new_node);
        assert(rv == 0);

        /*
         * Prepare a copy of the original board for simulating a match.
         */
        go_board_copy_board(work_board, orig_board);

        // Need setup coordinates of node before getValueByMC()
        new_node->x = move_array_xxx[j].x;
        new_node->y = move_array_xxx[j].y;
        const double value = getValueByMC(new_node, work_board, color);

        new_node->value = value;
        new_node->num_win = 0;
        if (value == 1.0) {
            new_node->num_win++;
        }
        new_node->num_match = 1;
        new_node->num_children = 0;
    }


#ifdef DIAGNOSTIC
    {
        int num_error = 0;
        for (struct uct_node* child = root->children; child != NULL;
             child = child->next) {
            bool is_confirmed = false;
            for (int i = 0; i < num_move; i++) {
                if ((child->x == move_array_xxx[i].x)
                    && (child->y == move_array_xxx[i].y)) {
                    is_confirmed = true;
                    break;
                }
            }
            if (!is_confirmed) {
                printf("Child node (%d,%d) is invalid.\n", child->x, child->y);
                num_error++;
            }
        }
        if (num_error > 0) {
            printf("After initialization.\n");
            printf("Error detected: %d error of %d children\n", num_error,
                   root->num_children);
            assert(num_error == 0);
        }
    }
#endif /* DIAGNOSTIC */


    /*
     * Simulate a lot of match
     */
    for (;;) {
        const int64_t current_time = current_time_get();
        if (current_time >= end_time) {
            break;
        }

        /*
         * Continue simulation even if this function returns an error.
         * On an error, this function just fails to create a new node.
         */
        playOneSequence(&g_uct_tree, end_time);
    }


    /*
     * Select the best child of root
     */
    struct uct_node* best_child;
    rv = uct_tree_get_max_child_with_limit(&g_uct_tree, root,
                                           UCT_CREATE_NODE_LIMIT, &best_child);
    if (rv != 0) {
        /*
         * Choose max child less than the limit.
         */
        rv = uct_tree_get_max_child(&g_uct_tree, root, &best_child);
        assert(rv == 0);
        assert(best_child != NULL);
        rv = 1; /* Bad solution */
    }

    const coord_t best_x = best_child->x;
    const coord_t best_y = best_child->y;
    const double best_score = best_child->value;

    /*
     * Check the candidate. If an eye is broken, pass this turn.
     * Such checks will be moved into go_engine_search_mc().
     */
    if (is_eye_position(orig_board, color, best_x, best_y)) {
        *x = STONE_COORD_INVALID;
        *y = STONE_COORD_INVALID;
        return -2; /* PASS */
    }

    *x = best_x;
    *y = best_y;
    *score = best_score;


#ifdef DIAGNOSTIC
    {
        int num_error = 0;
        for (struct uct_node* child = root->children; child != NULL;
             child = child->next) {
            bool is_confirmed = false;
            for (int i = 0; i < num_move; i++) {
                if ((child->x == move_array_xxx[i].x)
                    && (child->y == move_array_xxx[i].y)) {
                    is_confirmed = true;
                    break;
                }
            }
            if (!is_confirmed) {
                printf("Child node (%d,%d) is invalid.\n", child->x, child->y);
                num_error++;
            }
        }
        if (num_error > 0) {
            printf("Before termination.\n");
            printf("Error detected: %d error of %d children\n", num_error,
                   root->num_children);
            assert(num_error == 0);
        }
    }
#endif /* DIAGNOSTIC */


#ifdef UCT_DEBUG
    uct_tree_dump_tree(&g_uct_tree, g_uct_tree.root);
#endif /* UCT_DEBUG */


    /*
     * Cleanup tree for UCT
     */
    uct_tree_terminate(&g_uct_tree);


#ifdef ENABLE_PLAYOUT_PROFILE
    const int64_t t2 = current_time_get();

    /*
     * Display profile
     */
    printf("%s: Playout prifle\n", __FUNCTION__);
    printf("# of playouts : %d\n", num_playout);
    printf("Execution time: %f(sec)\n", ((double)(t2 - t1))/(1000*1000));
    printf("Win rate      : %f\n", *score);
    printf("Num match     : %d\n", best_child->num_match);
#endif /* ENABLE_PLAYOUT_PROFILE */


    assert((rv == 0) || (rv == 1));
    return rv; /* 0: Well simulated, 1: Not so simulated */
}


/***
 ***
 *** Internal functions
 ***
 ***/

/**
 * This function walks through the minmax tree by using USB1,
 * and add a new node as a leaf.
 *
 * If a valid move is found after the walkthrough and this is added
 * as a leaf node of the minmax tree, this function returns 0.
 * Otherwise it returns a negative value.
 *
 * \param minmax_tree  Minmax tree
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int playOneSequence(struct uct_tree* minmax_tree, int64_t end_time)
{
    int rv;

    struct go_board* orig_board = go_engine_get_main_board();
    struct go_board* work_board = go_engine_get_work_board();
#ifndef MEMLEAK_WORKAROUND
    struct go_board* temp_board = go_engine_get_work_board_rear();
#endif /* MEMLEAK_WORKAROUND */

    /*
     * Prepare a copy of the original board for simulating a match.
     */
    go_board_copy_board(work_board, orig_board);


    /*
     * Walktrough the UCT tree from the root to a leaf.
     */
    node_array[0] = minmax_tree->root;

    int i = 0;
    stone_t color = my_color_get();
    while (1) {
        assert(i < (NUM_NODE_ARRAY-1));

        if (node_array[i]->children == NULL) {
            break;
        }

        struct uct_node* next_node = descendByUCB1(node_array[i]);
        assert(next_node != NULL);

        /*
         * Place a stone to the position of next_node
         */
        stone_place(work_board, color, next_node->x, next_node->y);
        color = reverse_color_get(color);

        node_array[++i] = next_node; // node_array[i+1]
    }

    /*
     * Check whether node_array[i] is "pass" or not.
     * XXX: Skip this if node structure has "pass" or not.
     */
    int num_move;
    rv = valid_move_get(work_board, color, move_array, NUM_MOVE_ARRAY,
                        &num_move);
    if (rv != 0) {
        return -1;
    }

    /*
     * Create child nodes if node_array[i] has been visited enough.
     */
    struct uct_node* parent = node_array[i];
    if (parent->num_match < UCT_CREATE_NODE_LIMIT) {
        /*
         * Just perform a simulation.
         */
        const double value = getValueByMC(parent, work_board, color);
        parent->num_match++;
        if (value == 1.0) {
            parent->num_win++;
        }
        parent->value = ((double)(parent->num_win))/parent->num_match;

        /*
         * Update value of node_array[0], ..., node_array[i-1]
         */
        updateValue(node_array, i, 1 - value);
    } else {
        assert(parent != minmax_tree->root);

        /*
         * Add all the valid moves to the list of children.
         */
        for (int j = 0; j < num_move; j++) {
#ifndef MEMLEAK_WORKAROUND
            const int64_t current_time = current_time_get();
            if (current_time >= end_time) {
                break;
            }
#endif /* MEMLEAK_WORKAROUND */

            struct uct_node* new_node;
            rv = uct_tree_create_node(minmax_tree, &new_node);
            assert(rv == 0);
            rv = uct_tree_add_child(minmax_tree, parent, new_node);
            assert(rv == 0);

            // Need setup coordinates of node before getValueByMC()
            new_node->x = move_array[j].x;
            new_node->y = move_array[j].y;

            color = reverse_color_get(color);
#ifdef MEMLEAK_WORKAROUND
            const double value = 1.0;
#else  /* MEMLEAK_WORKAROUND */
            go_board_copy_board(temp_board, work_board);
            const double value = getValueByMC(new_node, temp_board, color);
#endif /* MEMLEAK_WORKAROUND */
            new_node->value = value;
            if (value == 1.0) {
                new_node->num_win = 1;
            } else {
                new_node->num_win = 0;
            }
            new_node->num_match = 1;

            /*
             * Update value of node_array[0], ..., node_array[i]
             */
            updateValue(node_array, i+1, 1 - value);
        }
    }

    return 0;
}


/**
 * This function searches the best move among the children of \a node
 * by using UCB1 method.
 *
 * Assume that \a nd must have at least one child.
 * Otherwise this function falls into the unexpected behavior.
 *
 * \param node  Node
 *
 * \retval node  The best move next to \a nd.
 */
struct uct_node* descendByUCB1(struct uct_node* node)
{
    assert(node->children != NULL);


    /*
     * Get the sum of num_match of all the children
     */
    int sum = 0;
    for (struct uct_node* child = node->children; child != NULL;
         child = child->next) {
        sum += child->num_match;
        assert((0.0 <= child->value) && (child->value <= 1.0));
    }


    /*
     * Search the child with the maximum win-rate.
     */
    double max_value = -HUGE_VAL; // -INFINITY
    struct uct_node* max_child = NULL;
    for (struct uct_node* child = node->children; child != NULL;
         child = child->next) {
        double value;
        if (child->num_match == 1) {
            value = HUGE_VAL; // INFINITY
        } else {
            assert(sum >= 1);
            assert(child->num_match > 0);

            value = 1.0 - child->value +
                sqrt(2 * log((double)sum) / child->num_match);
        }

        if (value > max_value) {
            max_value = value;
            max_child = child;
        }
    }
    assert(max_child != NULL);
    assert(max_child->value <= 1.0);


    return max_child;
}


/** \internal
 * This function updates nd_array[0], ..., nd_array[num-1].
 * The num_win, num_match and value of each node are updated.
 *
 * \a value is 1.0 or 0.0. It shows which stone wins in a playout.
 * If 1.0 is specified, the color of nd_array[num-1] wins the match.
 * Otherwise it loses the match.
 *
 * \param nd_array  Array of nodes
 * \param num       Number of nodes in the array
 * \param value     Value
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int updateValue(struct uct_node** nd_array, int num, double value)
{
    assert(num > 0);
    assert((value == 1.0) || (value == 0.0));

    double new_value = value;
    for (int i = num - 1; i >= 0; i--) {
        struct uct_node* nd = nd_array[i];

        if ((i % 2) == 1) { /* my turn */
            if (new_value == 1.0) {
                nd->num_win++;
            }
            nd->num_match++;
            nd->value = ((double)nd->num_win) / nd->num_match;
        } else { /* opponent's turn */
            if (new_value == 0.0) {
                nd->num_win++;
            }
            nd->num_match++;
            nd->value = 1.0 - ((double)nd->num_win) / nd->num_match;
        }
        assert((0.0 <= nd->value) && (nd->value <= 1.0));
        new_value = 1.0 - new_value;
    }

    return 0;
}


/**
 * This function gets the value of \a node by simulating the match.
 *
 * \param node      Node
 * \param color     Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 *
 * \retval 1.0  \a node wins the match by a simulation.
 * \retval 0.0  \a node loses the match by a simulation.
 */
double getValueByMC(struct uct_node* node, struct go_board* board,
                    stone_t color)
{
#ifdef DIAGNOSTIC
    {
        const int BOARD_SIZE = go_engine_get_board_size();

        if ((node->x < 0) || (node->y < 0) || (BOARD_SIZE <= node->x)
            || (BOARD_SIZE <= node->y)) {
            printf("%s: Given position is wrong. (%d,%d)\n", __FUNCTION__,
                   node->x, node->y);
            abort();
        }
    }
#endif /* DIAGNOSTIC */

    /*
     * Examine \a nd by a playout
     */
    double result;
    mc_playout(board, color, node->x, node->y, &result);
    assert((result == 1.0) || (result == 0.0));

#ifdef ENABLE_PLAYOUT_PROFILE
    num_playout++;
#endif /* ENABLE_PLAYOUT_PROFILE */

    return result;
}


/** \internal
 * This function simulates a match ("playout") and returns its result.
 *
 * \param board   Current game board
 * \param x       Coordinate of x-axis of the next move
 * \param y       Coordinate of y-axis of the next move
 * \param result  (OUT) Result of the match (Win: 1.0, Lose: 0.0)
 *
 * \retval 0   Success
 */
int mc_playout(struct go_board* board, stone_t color, coord_t x, coord_t y,
               double* result)
{
    int rv;
    const stone_t my_color = color;
    const stone_t op_color = reverse_color_get(my_color);

#ifdef DIAGNOSTIC
    {
        const int BOARD_SIZE = go_engine_get_board_size();

        if ((x < 0) || (y < 0) || (BOARD_SIZE <= x) || (BOARD_SIZE <= y)) {
            printf("%s: Given position is wrong. (%d,%d)\n", __FUNCTION__,
                   x, y);
            abort();
        }
    }
#endif /* DIAGNOSTIC */

    /*
     * The 1st move for my turn
     */
    stone_place(board, my_color, x, y);

    /*
     * Each player places a stone in turn randomly.
     * If there are no valid spaces left, terminate the loop.
     */
    bool is_passed = false;
    for (int i = 1; i < MAX_PLAYOUT_MOVE_NUM; i++) {
        stone_t color;
        if ((i % 2) != 0) { /* opponent's turn */
            color = op_color;
        } else { /* my turn */
            color = my_color;
        }

        /*
         * Get a candidate move
         */
        coord_t cx, cy;
        rv = mc_get_candidate_move(board, color, &cx, &cy);
        if (rv == 0) {
            is_passed = false;
            stone_place(board, color, cx, cy);
        } else {
            /*
             * No candidates. Need a pass.
             */
            if (is_passed) {
                break;
            } else {
                is_passed = true;
                continue;
            }
        }
    }

    /*
     * Get the result. (Win: 1.0, Lose: 0.0)
     */
    *result = mc_get_match_result(board, my_color);

    return 0;
}


static coord2d_t vm_array[MAX_BOARD_SIZE * MAX_BOARD_SIZE];
static coord2d_t cm_array[MAX_BOARD_SIZE * MAX_BOARD_SIZE];


/**
 * This function returns a valid candidate move.
 *
 * \param board   Current game board
 * \param color   Stone color
 * \param cx      (OUT) X coordinate of the move
 * \param cy      (OUT) Y coordinate of the move
 *
 * \retval 0   Success
 * \retval -1  No valid move (Should pass the turn.)
 */
int mc_get_candidate_move(struct go_board* board, stone_t color, coord_t* cx,
                          coord_t* cy)
{
    int rv;

    const int BOARD_SIZE = go_engine_get_board_size();
    const int num_array = BOARD_SIZE * BOARD_SIZE;

    /*
     * Get valid candidates
     */
    int num_move;
    rv = valid_move_get(board, color, vm_array, num_array, &num_move);
    if (rv != 0) {
        return -1;
    }

#ifdef DIAGNOSTIC
    assert(num_move > 0);

    for (int i = 0; i < num_move; i++) {
        if ((vm_array[i].x < 0) || (vm_array[i].y < 0)
            || (vm_array[i].x >= BOARD_SIZE)
            || (vm_array[i].y >= BOARD_SIZE)) {
            printf("%s: invalid node (%d,%d) at idx %d\n", __FUNCTION__,
                   vm_array[i].x, vm_array[i].y, i);
        }
    }
#endif /* DIAGNOSTIC */

    /*
     * Rule out a move from the candidates if it breaks an eye.
     * If there are no candidates left, "pass" is returned.
     */
    int num_cand = 0;
    for (int i = 0; i < num_move; i++) {
        if (!is_eye_position(board, color, vm_array[i].x, vm_array[i].y)) {
            cm_array[i].x = vm_array[i].x;
            cm_array[i].y = vm_array[i].y;
            num_cand++;
        }
    }
    if (num_cand == 0) {
        return -1;
    }

    /*
     * Choose a move randomly from cm_array[] which has better candidates.
     */
    const int idx = mc_get_random_number(0, num_cand-1);

    *cx = cm_array[idx].x;
    *cy = cm_array[idx].y;

    return 0;
}


/** \internal
 * This function evaluates the match result of the current board.
 *
 * \param board  Current game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 *
 * \retval 1.0  \a color wins the match.
 * \retval 0.0  \a color loses the match.
 */
float mc_get_match_result(struct go_board* board, stone_t color)
{
    /*
     * Calculate the number of stones on the board.
     */
    const int num_black = go_board_get_black(board);
    const int num_white = go_board_get_white(board);

    /*
     * Calculate eyes fo each stone.
     */
    const int BOARD_SIZE = go_engine_get_board_size();

    int num_black_territory = 0;
    int num_white_territory = 0;

    for (coord_t y = 0; y < BOARD_SIZE; y++) {
        for (coord_t x = 0; x < BOARD_SIZE; x++) {
            if (is_eye_position(board, STONE_COLOR_BLACK, x, y)) {
                num_black_territory++;
                continue;
            }
            if (is_eye_position(board, STONE_COLOR_WHITE, x, y)) {
                num_white_territory++;
                continue;
            }
        }
    }

    /*
     * Calculate the total score.
     */
    float score = (float)(num_black + num_black_territory - num_white
                          - num_white_territory);
    if (color == STONE_COLOR_BLACK) {
        score -= go_engine_get_komi();
    }

    if (color == STONE_COLOR_BLACK) {
        if (score > 0.0) {
            return 1.0; // Win
        } else {
            return 0.0; // Lose
        }
    } else {
        if (score < 0.0) {
            return 1.0; // Win
        } else {
            return 0.0; // Lose
        }
    }
}


/** \internal
 * This function returns a random number within [\a start, \a end].
 *
 * \param start  Start number of the interval
 * \param end    End number of the internal
 *
 * \retval random_num  Random number
 *
 * \todo Use better function than rand().
 */
int mc_get_random_number(int start, int end)
{
    if (start == end) {
        return start;
    } else {
        return (start + rand() % (end - start + 1));
    }
}
