#include "go_board.h"
#include "go_engine.h"
#include "go_engine_internal.h"
#include "go_string.h"

#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>


/*
 * Imported functions from go_engine.c
 */
extern int go_engine_get_board_size(void);

extern stone_t my_color_get(void);
extern stone_t opponent_color_get(void);
extern stone_t reverse_color_get(stone_t);

/*
 * Internal functions
 */
static bool is_empty_position(struct go_board*, coord_t, coord_t);
static bool is_all_same_color(struct go_board*, stone_t);
static bool is_surrounded_position_single(struct go_board*, stone_t, coord_t,
                                          coord_t);


/** \internal
 * This function places a stone (\a color) to the given position.
 *
 * \param board  Current game board
 * \parma color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval 0   Success
 * \retval -1  The given move is invalid.
 *
 * \note This function uses string functions.
 */
int stone_place(struct go_board* board, stone_t color, coord_t x, coord_t y)
{
    int rv;
    stone_t rev_color = reverse_color_get(color);
    const int BOARD_SIZE = go_engine_get_board_size();

    if (!is_valid_move(board, color, x, y)) {
        return -1;
    }
    assert(is_empty_position(board, x, y));

    /*
     * If the move satisfies the "ko" condition, memorise the position of
     * the stone which will be captured.
     */
    coord_t cx, cy;
    if (is_ko_move(board, color, x, y, &cx, &cy)) {
        go_board_set_ko_position(board, cx, cy);
    } else {
        go_board_set_ko_position(board, STONE_COORD_INVALID,
                                 STONE_COORD_INVALID);
    }

    go_board_set_stone(board, color, x, y);

    {
        int idx = 0;
        string_id_t checked[4];
        struct go_string_list* lst = board->string_list;

        if (x > 0) {
            const string_id_t next_id = go_string_list_get_id(lst, x-1, y);
            if (next_id != STRING_ID_INVALID) {
                stone_t str_color;
                int str_size;
                int str_liberty;
                go_string_list_get_info2(lst, next_id, &str_color, &str_size,
                                         &str_liberty);
                if (str_color == color) {
                    go_string_list_add_stone(lst, color, x, y, next_id);
                    checked[idx++] = next_id;
                }
            }
        }
        if (x < (BOARD_SIZE - 1)) {
            const string_id_t next_id = go_string_list_get_id(lst, x+1, y);
            bool is_checked = false;
            for (int i = 0; i < idx; i++) {
                if (checked[i] == next_id) {
                    is_checked = true;
                    break;
                }
            }
            if ((next_id != STRING_ID_INVALID) && !is_checked) {
                stone_t str_color;
                int str_size;
                int str_liberty;
                go_string_list_get_info2(lst, next_id, &str_color, &str_size,
                                         &str_liberty);
                if (str_color == color) {
                    if (idx == 0) {
                        go_string_list_add_stone(lst, color, x, y, next_id);
                    } else {
                        const string_id_t my_id =
                            go_string_list_get_id(lst, x, y);
                        go_string_list_merge_string(lst, next_id, my_id);
                    }
                    checked[idx++] = next_id;
                }
            }
        }
        if (y > 0) {
            const string_id_t next_id = go_string_list_get_id(lst, x, y-1);
            bool is_checked = false;
            for (int i = 0; i < idx; i++) {
                if (checked[i] == next_id) {
                    is_checked = true;
                    break;
                }
            }
            if ((next_id != STRING_ID_INVALID) && !is_checked) {
                stone_t str_color;
                int str_size;
                int str_liberty;
                go_string_list_get_info2(lst, next_id, &str_color, &str_size,
                                         &str_liberty);
                if (str_color == color) {
                    if (idx == 0) {
                        go_string_list_add_stone(lst, color, x, y, next_id);
                    } else {
                        const string_id_t my_id =
                            go_string_list_get_id(lst, x, y);
                        go_string_list_merge_string(lst, next_id, my_id);
                    }
                    checked[idx++] = next_id;
                }
            }
        }
        if (y < (BOARD_SIZE - 1)) {
            const string_id_t next_id = go_string_list_get_id(lst, x, y+1);
            bool is_checked = false;
            for (int i = 0; i < idx; i++) {
                if (checked[i] == next_id) {
                    is_checked = true;
                    break;
                }
            }
            if ((next_id != STRING_ID_INVALID) && !is_checked) {
                stone_t str_color;
                int str_size;
                int str_liberty;
                go_string_list_get_info2(lst, next_id, &str_color, &str_size,
                                         &str_liberty);
                if (str_color == color) {
                    if (idx == 0) {
                        go_string_list_add_stone(lst, color, x, y, next_id);
                    } else {
                        const string_id_t my_id =
                            go_string_list_get_id(lst, x, y);
                        go_string_list_merge_string(lst, next_id, my_id);
                    }
                    checked[idx++] = next_id;
                }
            }
        }
        if (idx == 0) {
            string_id_t my_id;
            go_string_list_create_string(lst, color, x, y, &my_id);
        }
    }

    /*
     * Check whther the opponent's stones are captured or not.
     * If yes, remove the captured stones.
     */
    if (x > 0) {
        if (is_surrounded_position(board, rev_color, x-1, y)) {
            rv = string_remove(board, rev_color, x-1, y);
            assert(rv > 0);
        }
    }
    if (x < (BOARD_SIZE - 1)) {
        if (is_surrounded_position(board, rev_color, x+1, y)) {
            rv = string_remove(board, rev_color, x+1, y);
            assert(rv > 0);
        }
    }
    if (y > 0) {
        if (is_surrounded_position(board, rev_color, x, y-1)) {
            rv = string_remove(board, rev_color, x, y-1);
            assert(rv > 0);
        }
    }
    if (y < (BOARD_SIZE - 1)) {
        if (is_surrounded_position(board, rev_color, x, y+1)) {
            rv = string_remove(board, rev_color, x, y+1);
            assert(rv > 0);
        }
    }

    return 0;
}


/* XXX */
#define ARRAY_SIZE (MAX_BOARD_SIZE * MAX_BOARD_SIZE)
static coord2d_t stone_array[ARRAY_SIZE];

/** \internal
 * This function removes a group of stones (as known as "string") which are
 * connected to the given stone.
 *
 * \param board  Current game board
 * \param color  Stone color
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval num_stones  Number of stones removed by this function
 *
 * \note This function uses string function.
 */
int string_remove(struct go_board* board, stone_t color, coord_t x, coord_t y)
{
    int rv, num;

    assert(go_board_get_stone(board, x, y) == color);
    assert((color == STONE_COLOR_BLACK) || (color == STONE_COLOR_WHITE));

    rv = go_string_list_get_string_stone(board->string_list, color, x, y,
                                         ARRAY_SIZE, stone_array, &num);
    if (rv != 0) {
        return 0;
    }

    for (int i = 0; i < num; i++) {
        go_board_remove_stone(board, stone_array[i].x, stone_array[i].y);
    }

    const string_id_t id = go_string_list_get_id(board->string_list, x, y);
    rv = go_string_list_destroy_string(board->string_list, id);
    assert(rv == 0);

    return num;
}


/** \internal
 * This function tells whether the given move is valid or not.
 *
 * The move is specified by \a color, \a x and \a y.
 * The color of the stone is \a color.
 * The position of the stone is \a x and \a y.
 *
 * \param board  Current game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval true   The given move is valid.
 * \retval false  The given move is invalid.
 *
 * \note This function uses string functions indirectly.
 */
bool is_valid_move(struct go_board* board, stone_t color, coord_t x, coord_t y)
{
    /*
     * Empty position?
     */
    if (!is_empty_position(board, x, y)) {
        return false;
    }

    /*
     * Suicide move?
     */
    if (is_suicide_move(board, color, x, y)) {
        return false;
    }

    /*
     * Ko?
     */
    coord_t cx, cy;
    if (is_ko_move(board, color, x, y, &cx, &cy)) {
        coord_t ko_x, ko_y;
        go_board_get_ko_position(board, &ko_x, &ko_y);

        if ((ko_x == x) && (ko_y == y)) {
            return false;
        }
    }

    return true;
}


/** \internal
 * This function tells whether the given move is a suicide move or not.
 *
 * The move is specified by \a color, \a x and \a y.
 * The color of the stone is \a color.
 * The position of the stone is \a x and \a y.
 *
 * \param board  Current game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval true   The given move is a suicide move.
 * \retval false  The given move is not a suicide move.
 *
 * \note This function uses string functions.
 */
bool is_suicide_move(struct go_board* board, stone_t color, coord_t x,
		     coord_t y)
{
    int rv, num;

    assert(is_empty_position(board, x, y));

    /*
     * Condition of suicide move:
     * (1) (x,y) is an empty point.
     * (2) All the neighbor positions have stones. (not empty)
     *       - If either neighbor position is empty, it is NOT a suicide move.
     * (3) For all the neighbor strings with the reverse color,
     *     the number of liberty points is more than one.
     *       - If the number of liberty point is one, the string is captured.
     *         In such case, it is NOT a suicide move.
     * (4) For all the neighbor strings with the same color, the number of
     *     liberty point is one.
     *       - If the number of liberty points of either string is
     *         more than one, the (merged) string cannot be captured.
     *         In such case, it is NOT a suicide move.
     */
    const int BOARD_SIZE = go_engine_get_board_size();
    const stone_t rev_color = reverse_color_get(color);

    if (x > 0) { /* left */
        const stone_t nc = go_board_get_stone(board, x-1, y);
        if (nc == STONE_COLOR_EMPTY) {
            return false;
        }

        rv = go_string_list_get_liberty_number(board->string_list, nc,
                                               x-1, y, &num);
        assert(rv == 0);

        if ((nc == color) && (num > 1)) {
            return false;
        }
        if ((nc != color) && (num == 1)) {
            assert(nc == rev_color);
            return false;
        }
    }
    if (x < (BOARD_SIZE - 1)) { /* right */
        const stone_t nc = go_board_get_stone(board, x+1, y);
        if (nc == STONE_COLOR_EMPTY) {
            return false;
        }

        rv = go_string_list_get_liberty_number(board->string_list, nc,
                                               x+1, y, &num);
        assert(rv == 0);

        if ((nc == color) && (num > 1)) {
            return false;
        }
        if ((nc != color) && (num == 1)) {
            assert(nc == rev_color);
            return false;
        }
    }
    if (y > 0) { /* down */
        const stone_t nc = go_board_get_stone(board, x, y-1);
        if (nc == STONE_COLOR_EMPTY) {
            return false;
        }

        rv = go_string_list_get_liberty_number(board->string_list, nc,
                                               x, y-1, &num);
        assert(rv == 0);

        if ((nc == color) && (num > 1)) {
            return false;
        }
        if ((nc != color) && (num == 1)) {
            assert(nc == rev_color);
            return false;
        }
    }
    if (y < (BOARD_SIZE - 1)) { /* up */
        const stone_t nc = go_board_get_stone(board, x, y+1);
        if (nc == STONE_COLOR_EMPTY) {
            return false;
        }

        rv = go_string_list_get_liberty_number(board->string_list, nc,
                                               x, y+1, &num);
        assert(rv == 0);

        if ((nc == color) && (num > 1)) {
            return false;
        }
        if ((nc != color) && (num == 1)) {
            assert(nc == rev_color);
            return false;
        }
    }

    return true;
}


/** \internal
 * This function tells whether the given move is a ko move or not.
 *
 * The move is specified by \a color, \a x and \a y.
 * The color of the stone is \a color.
 * The position of the stone is \a x and \a y.
 *
 * A "ko move" satisfies the following conditions:
 *     (1) The given position is empty. It is OK to place a stone.
 *     (2) All the adjacent stones have the reverse color of the given stone.
 *     (3) Placing a stone captures only one stone.
 *
 * When is_ko_move() returns true, the board becomes "ko" by committing
 * the move. In the next move, the opponent cannot place a stone to
 * the position (\a cx, \a cy) where a stone is captured.
 *
 * \param board  Current game board
 * \param color  Stone color
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 * \param cx     (OUT) Coordinate of x-axis for captured stone
 * \param cy     (OUT) Coordinate of y-axis for captured stone
 *
 * \retval true   The given move is a ko move.
 * \retval false  The given move is not a ko move.
 */
bool is_ko_move(struct go_board* board, stone_t color, coord_t x, coord_t y,
		coord_t* cx, coord_t* cy)
{
    const stone_t rev_color = reverse_color_get(color);
    const int BOARD_SIZE = go_engine_get_board_size();

    assert(is_empty_position(board, x, y));

    /*
     * Check whthether the given position is surrounded by the opponent's
     * stones (or the boundaries) or not.
     */
    if (x > 0) {
        if (go_board_get_stone(board, x-1, y) != rev_color) {
            return false;
        }
    }
    if (x < (BOARD_SIZE - 1)) {
        if (go_board_get_stone(board, x+1, y) != rev_color) {
            return false;
        }
    }
    if (y > 0) {
        if (go_board_get_stone(board, x, y-1) != rev_color) {
            return false;
        }
    }
    if (y < (BOARD_SIZE - 1)) {
        if (go_board_get_stone(board, x, y+1) != rev_color) {
            return false;
        }
    }

    /*
     * If the given position is surrounded by the opponent's stones,
     * check whether the place can capture only one stone or not.
     */
    {
        int num_captured = 0;

        go_board_set_stone(board, color, x, y); /* tentatively */

        if (x > 0) { /* (x-1, y) */
            if (is_surrounded_position_single(board, rev_color, x-1, y)) {
                num_captured++;
                *cx = x-1;
                *cy = y;
            }
        }
        if (x < (BOARD_SIZE - 1)) { /* (x+1, y) */
            if (is_surrounded_position_single(board, rev_color, x+1, y)) {
                num_captured++;
                *cx = x+1;
                *cy = y;
            }
        }
        if (y > 0) { /* (x, y-1) */
            if (is_surrounded_position_single(board, rev_color, x, y-1)) {
                num_captured++;
                *cx = x;
                *cy = y-1;
            }
        }
        if (y < (BOARD_SIZE - 1)) { /* (x, y+1) */
            if (is_surrounded_position_single(board, rev_color, x, y+1)) {
                num_captured++;
                *cx = x;
                *cy = y+1;
            }
        }

        go_board_remove_stone(board, x, y);

        if (num_captured != 1) {
            return false;
        }
    }

    return true;
}


/** \internal
 * This function tells whether all the adjacent stones of the given position
 * are the reverse color of \a color or not.
 *
 * \param board  Current game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval true   The given move is surrounded.
 * \retval false  The given move is not surrounded.
 */
bool is_surrounded_position_single(struct go_board* board, stone_t color,
				   coord_t x, coord_t y)
{
    const stone_t rev_color = reverse_color_get(color);
    const int BOARD_SIZE = go_engine_get_board_size();
    bool is_surrounded = true;

    if (x > 0) {
        if (go_board_get_stone(board, x-1, y) != rev_color) {
            is_surrounded = false;
        }
    }
    if (x < BOARD_SIZE) {
        if (go_board_get_stone(board, x+1, y) != rev_color) {
            is_surrounded = false;
        }
    }
    if (y > 0) {
        if (go_board_get_stone(board, x, y-1) != rev_color) {
            is_surrounded = false;
        }
    }
    if (y < BOARD_SIZE) {
        if (go_board_get_stone(board, x, y+1) != rev_color) {
            is_surrounded = false;
        }
    }

    return is_surrounded;
}


/** \internal
 * This function tells whether the given position is surrounded by the
 * stones with the reverse color or not.
 *
 * The target position is specified by \a color, \a x and \a y.
 * \a color is the color of stone. \a x and \a y is the coordinate.
 *
 * Actually a string to which the given stone belongs is checked.
 * If the string is surrounded by the reverse stones, this function returns
 * true. Otherwise it returns false.
 *
 * Caution that if the stone of the given position is not the same as
 * \a color, this function returns false.
 *
 * \param board  Current game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval true   The given stone is surrounded by the reverse stones.
 * \retval false  The given stone is not surrounded by the reverse stones.
 *
 * \note This function uses string functions.
 */
bool is_surrounded_position(struct go_board* board, stone_t color, coord_t x,
                            coord_t y)
{
    int rv, num;

    /*
     * Sanity check
     */
    if (go_board_get_stone(board, x, y) != color) {
        return false;
    }

    /*
     * Exclude the case where all the stones occupies the board.
     * Although all the stones are surrounded by the boundaries,
     * it should be excluded from "surrounded" stones.
     */
    if (is_all_same_color(board, color)) {
        return false;
    }

    rv = go_string_list_get_liberty_number(board->string_list, color, x, y,
                                           &num);
    if (rv != 0) {
        /* The given string is not found. */
        return false;
    }

    if (num == 0) {
        return true;
    } else {
        return false;
    }
}


/** \internal
 * This function gets an array of valid moves.
 *
 * \param board      Current game board
 * \param color      Stone color
 * \param array      (OUT) Array for valid moves
 * \param num_array  Number of elements of array
 * \param num_move   (OUT) Number of valid moves   
 *
 * \retval 0   Success
 * \retval -1  \a num_array is too small.
 * \retval -2  There are no valid moves.
 *
 * \note This function uses string functions indirectly.
 */
int valid_move_get(struct go_board* board, stone_t color, coord2d_t* array,
		   int num_array, int* num_move)
{
    const int BOARD_SIZE = go_engine_get_board_size();

    int num = 0;
    for (coord_t y = 0; y < BOARD_SIZE; y++) {
        for (coord_t x = 0; x < BOARD_SIZE; x++) {
            if (is_valid_move(board, color, x, y)) {
                array[num].x = x;
                array[num].y = y;
                num++;
            }
        }
    }

    /*
     * Array is too small.
     */
    if (num > num_array) {
        return -1;
    }

    if (num == 0) {
        *num_move = 0;
        return -2;
    } else {
        *num_move = num;
        return 0;
    }
}


/** \internal
 * This function tells whether the given position is empty or not.
 *
 * \param board  Game board
 * \param x      Coordinate of x-axis
 * \param y      Coordinate of y-axis
 *
 * \retval true   The given position is empty.
 * \retval false  The given position is not empty.
 */
bool is_empty_position(struct go_board* board, coord_t x, coord_t y)
{
    if (go_board_get_stone(board, x, y) == STONE_COLOR_EMPTY) {
	return true;
    } else {
	return false;
    }
}


/** \internal
 * This function tells whether all the stones on the board is the same as
 * \a color or not. If the board is filled with \a color, this function
 * returns true. Otherwise it returns false.
 *
 * \param board  Game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 *
 * \retval true   The board is filled with \a color.
 * \retval false  The board is not filled with \a color.
 */
bool is_all_same_color(struct go_board* board, stone_t color)
{
    const int BOARD_SIZE = go_engine_get_board_size();

    for (coord_t y = 0; y < BOARD_SIZE; y++) {
	for (coord_t x = 0; x < BOARD_SIZE; x++) {
	    if (go_board_get_stone(board, x, y) != color) {
		return false;
	    }
	}
    }

    return true;
}


/** \internal
 * This function tells whether the given position is an eye of stones
 * whose color is \a color.
 *
 * \param board  Game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Position (X-coordinate)
 * \param y      Position (Y-coordinate)
 *
 * \retval true   The given position is an eye of \a color.
 * \retval false  The given position is not an eye of \a color.
 *
 * \note This function uses string functions.
 */
bool is_eye_position(struct go_board* board, stone_t color, coord_t x,
                     coord_t y)
{
    int rv, num;

    const int BOARD_SIZE = go_engine_get_board_size();

    if (go_board_get_stone(board, x, y) != STONE_COLOR_EMPTY) {
        return false;
    }

    /*
     * (x, y) is an eye of the specified stone if all the conditions hold:
     *     - All the neighbors of (x, y) are \a color or "wall".
     *     - Each neighbor strings has move than one liberty points.
     */
    if (x > 0) {
        rv =  go_string_list_get_liberty_number(board->string_list, color,
                                                x-1, y, &num);
        if (rv != 0) {
            return false;
        }
        if (num < 2) {
            return false;
        }
    }
    if (x < (BOARD_SIZE - 1)) {
        rv =  go_string_list_get_liberty_number(board->string_list, color,
                                                x+1, y, &num);
        if (rv != 0) {
            return false;
        }
        if (num < 2) {
            return false;
        }
    }
    if (y > 0) {
        rv =  go_string_list_get_liberty_number(board->string_list, color,
                                                x, y-1, &num);
        if (rv != 0) {
            return false;
        }
        if (num < 2) {
            return false;
        }
    }
    if (y < (BOARD_SIZE - 1)) {
        rv =  go_string_list_get_liberty_number(board->string_list, color,
                                                x, y+1, &num);
        if (rv != 0) {
            return false;
        }
        if (num < 2) {
            return false;
        }
    }

    return true;
}


/** \internal
 * This function tells whether the given position is an disconnected
 * position for \a color.
 *
 * \param board  Game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Position (X-coordinate)
 * \param y      Position (Y-coordinate)
 *
 * \retval true   The given position is a disconnected position.
 * \retval false  The given position is not a disconnected position.
 */
bool is_disconnected_position(struct go_board* board, stone_t color,
                              coord_t x, coord_t y)
{
    const int BOARD_SIZE = go_engine_get_board_size();

    if (go_board_get_stone(board, x, y) != STONE_COLOR_EMPTY) {
        return false;
    }
    if ((x == 0) || (x == (BOARD_SIZE-1)) || (y == 0)
        || (y == (BOARD_SIZE-1))) {
        return false;
    }

    /*
     * Does the position match the following position?
     *   (1) String A and B are the same color as \a color.
     *   (2) Both are disconnected.
     *   (3) No stones with the same color exist in the other two direction.
     *
     *    X + X
     *    + # * <- String A
     *    X * +
     *      ^-- String B
     */
    if ((go_board_get_stone(board, x-1, y) == color)
        && (go_board_get_stone(board, x, y-1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x-1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y-1);
        if ((go_board_get_stone(board, x+1, y) != color)
            && (go_board_get_stone(board, x, y+1) != color) && (id1 != id2)) {
            return true;
        }
    }
    if ((go_board_get_stone(board, x-1, y) == color)
        && (go_board_get_stone(board, x, y+1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x-1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y+1);
        if ((go_board_get_stone(board, x+1, y) != color)
            && (go_board_get_stone(board, x, y-1) != color) && (id1 != id2)) {
            return true;
        }
    }
    if ((go_board_get_stone(board, x+1, y) == color)
        && (go_board_get_stone(board, x, y-1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x+1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y-1);
        if ((go_board_get_stone(board, x-1, y) != color)
            && (go_board_get_stone(board, x, y+1) != color) && (id1 != id2)) {
            return true;
        }
    }
    if ((go_board_get_stone(board, x+1, y) == color)
        && (go_board_get_stone(board, x, y+1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x+1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y+1);
        if ((go_board_get_stone(board, x-1, y) != color)
            && (go_board_get_stone(board, x, y-1) != color) && (id1 != id2)) {
            return true;
        }
    }

    return false;
}


/** \internal
 * This function tells whether the given position is an peeped
 * position for \a color.
 *
 * \param board  Game board
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Position (X-coordinate)
 * \param y      Position (Y-coordinate)
 *
 * \retval true   The given position is a peeped position.
 * \retval false  The given position is not a peeped position.
 */
bool is_peeped_position(struct go_board* board, stone_t color, coord_t x,
                        coord_t y)
{
    const int BOARD_SIZE = go_engine_get_board_size();
    const stone_t op_color = reverse_color_get(color);

    if (go_board_get_stone(board, x, y) != STONE_COLOR_EMPTY) {
        return false;
    }
    if ((x == 0) || (x == (BOARD_SIZE-1)) || (y == 0)
        || (y == (BOARD_SIZE-1))) {
        return false;
    }

    /*
     * Does the position match the following position?
     *   (1) String A and B are the same color as \a color.
     *   (2) Both are disconnected and exist in the same line/row.
     *   (3) One or two opponent stones exist in the other two direction.
     *
     *    X O X
     *    * # * <- String A (*)
     *    X + X
     *    ^-- String B (*)
     */
    if ((go_board_get_stone(board, x-1, y) == color)
        && (go_board_get_stone(board, x, y-1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x-1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y-1);
        if (((go_board_get_stone(board, x+1, y) == op_color)
             || (go_board_get_stone(board, x, y+1) == op_color))
            && (id1 != id2)) {
            return true;
        }
    }
    if ((go_board_get_stone(board, x-1, y) == color)
        && (go_board_get_stone(board, x, y+1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x-1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y+1);
        if (((go_board_get_stone(board, x+1, y) == op_color)
             || (go_board_get_stone(board, x, y-1) == op_color))
            && (id1 != id2)) {
            return true;
        }
    }
    if ((go_board_get_stone(board, x+1, y) == color)
        && (go_board_get_stone(board, x, y-1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x+1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y-1);
        if (((go_board_get_stone(board, x-1, y) == op_color)
             || (go_board_get_stone(board, x, y+1) == op_color))
            && (id1 != id2)) {
            return true;
        }
    }
    if ((go_board_get_stone(board, x+1, y) == color)
        && (go_board_get_stone(board, x, y+1) == color)) {
        const string_id_t id1 =
            go_string_list_get_id(board->string_list, x+1, y);
        const string_id_t id2 =
            go_string_list_get_id(board->string_list, x, y+1);
        if (((go_board_get_stone(board, x-1, y) == op_color)
             || (go_board_get_stone(board, x, y-1) == op_color))
            && (id1 != id2)) {
            return true;
        }
    }

    return false;
}
