#include <assert.h>
#include <stddef.h>
#include <stdio.h>

#include "go_board.h"
#include "go_string.h"


//#define DEBUG_STRING
#ifdef DEBUG_STRING
#define dprintf printf
#else  /* DEBUG_STRING */
#define dprintf(...)
#endif /* DEBUG_STRING */


#define GET_BOARD_SIZE(lst) (lst)->board_size
#define GET_STRING_ID(lst, x, y) (lst)->board_array[(y)][(x)]
#define SET_STRING_ID(lst, x, y, id) (lst)->board_array[(y)][(x)]=(id);
#define GET_STRING_PTR(lst, id) &((lst)->string_list[(id)])


/*
 * Internal functions
 */
static int go_string_list_create_string_subr(struct go_string_list*,
                                             struct go_string*, coord_t,
                                             coord_t, string_id_t*, int*);
static int go_string_list_add_stone_subr(struct go_string_list*,
                                         struct go_string*, coord_t, coord_t,
                                         coord_t, coord_t, string_id_t*,
                                         int*);
static int go_string_list_count_liberty_number(struct go_string_list* lst,
                                               struct go_string* str);
static int go_string_list_count_liberty_number_subr(struct go_string_list*,
                                                    struct go_string*,
                                                    coord_t, coord_t, int);
static int go_string_list_initialize_string_list(struct go_string_list*);


/*
 * String ID allocation
 */
static int go_string_list_allocate_id(struct go_string_list* lst);
static int go_string_list_release_id(struct go_string_list* lst, int);


/**
 * Constructor
 *
 * \param sz  Board size
 */
int go_string_list_initialize(struct go_string_list* lst, int sz)
{
    const int BOARD_SIZE = sz;

    lst->board_size = sz;
    lst->num_string = 0;

    /* XXX: tentative */
    for (int i = 0; i < MAX_STRING_NUM; i++) {
        lst->id_bitmap_array[i] = false; /* true: busy, false: empty */
    }

    for (coord_t y = 0; y < BOARD_SIZE; y++) {
        for (coord_t x = 0; x < BOARD_SIZE; x++) {
            lst->board_array[y][x] = STRING_ID_INVALID;
        }
    }

    go_string_list_initialize_string_list(lst);

    return 0;
}


/**
 * Destructor
 */
int go_string_list_terminate(struct go_string_list* lst)
{
    return 0;
}


/**
 * This function copies the string list status of \a src to \a dst.
 *
 * \param \a dst  String list which is the destination of copy
 * \param \a src  String list which is the source of copy
 *
 * \retval 0  Success
 */
int go_string_list_copy_list(struct go_string_list* dst,
                             struct go_string_list* src)
{
    dst->board_size = src->board_size;
    dst->num_string = src->num_string;

    /* XXX: tentative */
    for (int i = 0; i < MAX_STRING_NUM; i++) {
        dst->id_bitmap_array[i] = src->id_bitmap_array[i];
    }

    /*
     * XXX: Can omit copy if its ID is invalid.
     *      Also can select fields which are actually used.
     */
    for (int i = 0; i < MAX_STRING_NUM; i++) {
        dst->string_list[i] = src->string_list[i];
    }

    for (int y = 0; y < MAX_BOARD_SIZE; y++) {
        for (int x = 0; x < MAX_BOARD_SIZE; x++) {
            dst->board_array[y][x] = src->board_array[y][x];
        }
    }

    return 0;
}


/**
 * This function creates a new string with a stone.
 * The stone is specified by \a color, \a x and \a y.
 *
 * On success, a new string is created. This function returns 0.
 * An ID for the string is assigned and stored to \a id.
 * Otherwise this function returns a negative value.
 *
 * [What is updated after creation?]
 * (1) All the neighbor strings of the given stone
 *   - num_liberty is decreased by one.
 *   - num_adjacent is increased by one. (\a id is added.)
 *   - adjacent_id[] becomes larger. (\a id is added.)
 *
 * \param lst    String list
 * \param color  Stone color
 * \param x      Stone position (X-coordinate)
 * \param y      Stone position (Y-coordinate)
 * \param id     (OUT) String ID which is newly assigned
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_string_list_create_string(struct go_string_list* lst, stone_t color,
                                 coord_t x, coord_t y, string_id_t* id)
{
    assert(go_string_list_get_id(lst, x, y) == STRING_ID_INVALID);

    const int BOARD_SIZE = GET_BOARD_SIZE(lst);
    const string_id_t my_id = go_string_list_allocate_id(lst);
    struct go_string* str = GET_STRING_PTR(lst, my_id);

    assert((0 <= my_id) && (my_id < MAX_STRING_NUM));

    str->string_id = my_id;
    str->color = color;
    str->size = 1;
    str->num_liberty = 0;
    str->num_adjacent = 0;
    str->belong_id[0].x = x;
    str->belong_id[0].y = y;

    /*
     * Check neighbor positions of this string.
     * Update num_liberty, num_adjacent and adjacent_id[].
     */
    int num_checked = 0;
    string_id_t checked_list[4];
    if (x > 0) { /* left */
        go_string_list_create_string_subr(lst, str, x-1, y, checked_list,
                                          &num_checked);
    }
    if (x < (BOARD_SIZE - 1)) { /* right */
        go_string_list_create_string_subr(lst, str, x+1, y, checked_list,
                                          &num_checked);
    }
    if (y > 0) { /* down */
        go_string_list_create_string_subr(lst, str, x, y-1, checked_list,
                                          &num_checked);
    }
    if (y < (BOARD_SIZE - 1)) { /* up */
        go_string_list_create_string_subr(lst, str, x, y+1, checked_list,
                                          &num_checked);
    }

    lst->num_string++;
    SET_STRING_ID(lst, x, y, my_id);
    *id = my_id;

    return 0;
}


/** \internal
 * Sub function of go_string_list_create_string().
 * Check/update a string to which (\a nx, \a ny) belongs.
 */
int go_string_list_create_string_subr(struct go_string_list* lst,
                                      struct go_string* str, coord_t nx,
                                      coord_t ny, string_id_t* array,
                                      int* num_array)
{
    const string_id_t next_id = GET_STRING_ID(lst, nx, ny);
    if (next_id == STRING_ID_INVALID) {
        str->num_liberty++;
    } else {
        /*
         * If next_id is already counted, skip the update.
         */
        bool is_checked = false;
        for (int i = 0; i < *num_array; i++) {
            if (array[i] == next_id) {
                is_checked = true;
                break;
            }
        }

        if (!is_checked) {
            const string_id_t my_id = str->string_id;

            str->adjacent_id[str->num_adjacent] = next_id;
            str->num_adjacent++;

            struct go_string* nstr = GET_STRING_PTR(lst, next_id);
            nstr->num_liberty--;
            nstr->adjacent_id[nstr->num_adjacent] = my_id;
            nstr->num_adjacent++;

            array[*num_array] = next_id;
            (*num_array)++;
        }
    }

    return 0;
}


/**
 * This function adds a stone to an existing string.
 * The stone is specified by \a color, \a x and \a y.
 * The string is given by \a id. Both should be the same color.
 *
 * On success, this function returns 0.
 * The status of the given string and its neighbor strings is also updated.
 * Otherwise this function returns a negative value.
 *
 * [What is updated after addition?]
 * (1) String \a id
 *   - size increases by one.
 *   - belong_id[] increases by one entry. (The given stone is added.)
 *   - num_liberty may become larger. (Need calculation.)
 *   - num_adjacent may become larger. (Need calculation.)
 *   - adjacent_id[] may become larger. (Need calculation.)
 *
 * (2) All the neighbor strings of the given stone
 *   - num_adjacent is increased by one. (\a id is added.)
 *   - adjacent_id[] increases by one entry. (\a id is added.)
 *   - Note that the stat is unchanged if \a id is alerady included.
 *
 * \param lst    String list
 * \param color  Stone color
 * \param x      Stone position (X-coordinate)
 * \param y      Stone position (Y-coordinate)
 * \param id     String ID
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_string_list_add_stone(struct go_string_list* lst, stone_t color,
                             coord_t x, coord_t y, string_id_t id)
{
    assert(id != STRING_ID_INVALID);

    struct go_string* str = GET_STRING_PTR(lst, id);
    const int BOARD_SIZE = GET_BOARD_SIZE(lst);

    const int idx = str->size;
    str->size++;
    str->num_liberty--;
    str->belong_id[idx].x = x;
    str->belong_id[idx].y = y;

    /*
     * Check neighbor positions of this string.
     * Update num_liberty, num_adjacent and adjacent_id[].
     */
    int num_checked = 0;
    string_id_t checked_list[4];
    if (x > 0) { /* left */
        go_string_list_add_stone_subr(lst, str, x-1, y, x, y, checked_list,
                                      &num_checked);
    }
    if (x < (BOARD_SIZE - 1)) { /* right */
        go_string_list_add_stone_subr(lst, str, x+1, y, x, y, checked_list,
                                      &num_checked);
    }
    if (y > 0) { /* down */
        go_string_list_add_stone_subr(lst, str, x, y-1, x, y, checked_list,
                                      &num_checked);
    }
    if (y < (BOARD_SIZE - 1)) { /* up */
        go_string_list_add_stone_subr(lst, str, x, y+1, x, y, checked_list,
                                      &num_checked);
    }

    SET_STRING_ID(lst, x, y, id);

    return 0;
}


/** \internal
 * Sub function of go_string_list_add_stone().
 * Check/update a string to which (\a nx, \a ny) belongs.
 * (\a ox, \a oy) is the position of the added stone.
 */
int go_string_list_add_stone_subr(struct go_string_list* lst,
                                  struct go_string* str, coord_t nx,
                                  coord_t ny, coord_t ax, coord_t ay,
                                  string_id_t* array, int* num_array)
{
    const int BOARD_SIZE = GET_BOARD_SIZE(lst);
    const string_id_t my_id = str->string_id;

    const string_id_t next_id = GET_STRING_ID(lst, nx, ny);
    if (next_id != STRING_ID_INVALID) {
        bool is_counted = false;
        for (int i = 0; i < *num_array; i++) {
            if (array[i] == next_id) {
                is_counted = true;
                break;
            }
        }

        /*
         * Skip update if both are the same string.
         */
        if ((next_id != my_id) && !is_counted) {
            bool is_checked = false;
            const int num = str->num_adjacent;
            for (int i = 0; i < num; i++) {
                if (str->adjacent_id[i] == next_id) {
                    is_checked = true;
                    break;
                }
            }
            if (!is_checked) {
                str->adjacent_id[num] = next_id;
                str->num_adjacent++;
            }

            struct go_string* nstr = GET_STRING_PTR(lst, next_id);
            nstr->num_liberty--;

            is_checked = false;
            const int num_next = nstr->num_adjacent;
            for (int i = 0; i < num_next; i++) {
                if (nstr->adjacent_id[i] == my_id) {
                    is_checked = true;
                    break;
                }
            }
            if (!is_checked) {
                nstr->adjacent_id[num_next] = my_id;
                nstr->num_adjacent++;
            }

            array[*num_array] = next_id;
            (*num_array)++;
        }
    } else {
        /*
         * Check whether the empty point (nx, ny) is already counted or not.
         * Note that (ax, ay) should be ruled out in each case.
         */
        bool is_already_counted = false;
        if (nx > 0) { /* left */
            if (((nx-1) != ax) || (ny != ay)) {
                const string_id_t id = GET_STRING_ID(lst, nx-1, ny);
                if (my_id == id) {
                    is_already_counted = true;
                }
            }
        }
        if (nx < (BOARD_SIZE - 1)) { /* right */
            if (((nx+1) != ax) || (ny != ay)) {
                const string_id_t id = GET_STRING_ID(lst, nx+1, ny);
                if (my_id == id) {
                    is_already_counted = true;
                }
            }
        }
        if (ny > 0) { /* down */
            if ((nx != ax) || ((ny-1) != ay)) {
                const string_id_t id = GET_STRING_ID(lst, nx, ny-1);
                if (my_id == id) {
                    is_already_counted = true;
                }
            }
        }
        if (ny < (BOARD_SIZE - 1)) { /* up */
            if ((nx != ax) || ((ny+1) != ay)) {
                const string_id_t id = GET_STRING_ID(lst, nx, ny+1);
                if (my_id == id) {
                    is_already_counted = true;
                }
            }
        }
        if (!is_already_counted) {
            str->num_liberty++;
        }
    }

    return 0;
}


/**
 * This function merges two strings which are adjacent each other.
 * Both should be the same color.
 *
 * A string (\a id2) is merged into a string (\a id1).
 * After merge, all the stones of \a id2 belong to \a id1.
 *
 * On success, this function returns 0.
 * The status of the string and its neighbor strings is also updated.
 * Otherwise this function returns a negative value.
 *
 * [What is updated after merge?]
 * (1) String \a id1
 *   - size += (id2's size)
 *   - belong_id[] += (id2's belong_id[])
 *   - num_liberty += (id2's num_liberty - overlap)
 *   - num_adjacent becomes larger. (Intersections should be ruled out.)
 *   - adjacent_id[] becomes larger. (Same as the above.)
 *
 * (2) All the neighbor strings of \a id2
 *   - num_adjacent decreased by one because \a id2 disappears.
 *   - adjacent_id[] becomes smaller. (\a id2 is removed.)
 *
 * \param lst  String list
 * \param id1  String ID which continues
 * \param id2  String ID which disappears
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_string_list_merge_string(struct go_string_list* lst, string_id_t id1,
                                string_id_t id2)
{
    assert((id1 != STRING_ID_INVALID) && (id2 != STRING_ID_INVALID));

    struct go_string* str1 = GET_STRING_PTR(lst, id1);
    struct go_string* str2 = GET_STRING_PTR(lst, id2);

    assert(str1->color == str2->color);

    /*
     * Merge stones of str2 to str1.
     */
    for (int i = 0; i < str2->size; i++) {
        const int idx = str1->size;
        str1->belong_id[idx].x = str2->belong_id[i].x;
        str1->belong_id[idx].y = str2->belong_id[i].y;
        str1->size++;
    }

    /*
     * Update ID (id2 -> id1)
     */
    for (int i = 0; i < str2->size; i++) {
        const int x = str2->belong_id[i].x;
        const int y = str2->belong_id[i].y;
        assert(GET_STRING_ID(lst, x, y) == id2);
        SET_STRING_ID(lst, x, y, id1);
        lst->string_list[id2].string_id = STRING_ID_INVALID;
        assert(GET_STRING_ID(lst, x, y) == id1);
        /*
         * Update all the neighbor strings of str2. (except for str1)
         */
        for (int j = 0; j < str2->num_adjacent; j++) {
            const string_id_t id = str2->adjacent_id[j];
            if (id != id1) {
                struct go_string* str = GET_STRING_PTR(lst, id);
                /*
                 * If str has id1 in its adjacent list, just remove id2.
                 * Otherwise change id2 to id1.
                 */
                int idx = 0;
                bool is_present = false;
                const int num = str->num_adjacent;
                for (int k = 0; k < num; k++) {
                    if (str->adjacent_id[k] == id2) {
                        idx = k;
                    }
                    if (str->adjacent_id[k] == id1) {
                        is_present = true;
                    }
                }
                if (!is_present) {
                    str->adjacent_id[idx] = id1;
                    /* num_adjacent is unchanged because of no removal. */
                } else {
                    for (int k = idx; k < num-1; k++) {
                        str->adjacent_id[k] = str->adjacent_id[k+1];
                    }
                    str->num_adjacent--;
                }
            }
        }
    }
    go_string_list_release_id(lst, id2);

    /*
     * Recount liberty points.
     */
    str1->num_liberty = 0;
    go_string_list_count_liberty_number(lst, str1);

    /*
     * Remove id2 from adjacent_id[]. Then pack it.
     */
    int num = str1->num_adjacent;
    for (int i = 0; i < num; i++) {
        if (str1->adjacent_id[i] == id2) {
            for (int j = i; j < num - 1; j++) {
                str1->adjacent_id[j] = str1->adjacent_id[j+1];
            }
            str1->num_adjacent--;
            break;
        }
    }
    /*
     * Append str2->adjacent_id[] except for id1.
     * Also IDs which are included in str1->adjacent_id[] must be ruled out.
     */
    num = str1->num_adjacent;
    for (int i = 0; i < str2->num_adjacent; i++) {
        const string_id_t id = str2->adjacent_id[i];

        bool is_counted = false;
        for (int j = 0; j < num; j++) {
            if (id == str1->adjacent_id[j]) {
                is_counted = true;
                break;
            }
        }
        if ((id != id1) && !is_counted) {
            const int idx = str1->num_adjacent;
            str1->adjacent_id[idx] = str2->adjacent_id[i];
            str1->num_adjacent++;
        }
    }

    return 0;
}


/** \internal
 * This function counts the liberty points of the given string.
 */
int go_string_list_count_liberty_number(struct go_string_list* lst,
                                        struct go_string* str)
{
    const int BOARD_SIZE = GET_BOARD_SIZE(lst);
    const int NUM_STONE = str->size;

    for (int i = 0; i < NUM_STONE; i++) {
        const coord_t x = str->belong_id[i].x;
        const coord_t y = str->belong_id[i].y;

        if (x > 0) { /* left */
            go_string_list_count_liberty_number_subr(lst, str, x-1, y, i);
        }
        if (x < (BOARD_SIZE - 1)) { /* right */
            go_string_list_count_liberty_number_subr(lst, str, x+1, y, i);
        }
        if (y > 0) { /* down */
            go_string_list_count_liberty_number_subr(lst, str, x, y-1, i);
        }
        if (y < (BOARD_SIZE - 1)) { /* up */
            go_string_list_count_liberty_number_subr(lst, str, x, y+1, i);
        }
    }

    return 0;
}


/** \internal
 * Sub function of go_string_list_count_liberty_number().
 * This function counts the liberty points of the given string.
 */
int go_string_list_count_liberty_number_subr(struct go_string_list* lst,
                                             struct go_string* str, coord_t nx,
                                             coord_t ny, int idx)
{
    const int BOARD_SIZE = GET_BOARD_SIZE(lst);
    const string_id_t my_id = str->string_id;
    const coord_t x = str->belong_id[idx].x;
    const coord_t y = str->belong_id[idx].y;

    const string_id_t next_id = GET_STRING_ID(lst, nx, ny);
    if (next_id == STRING_ID_INVALID) {
        /*
         * Check whether the empty point (nx, ny) is already counted or not.
         * Note that (x, y) should be ruled out in each case.
         * Also (nx, ny) is ignored if it is already counted by one of
         * belong_id[0] ... belong_id[idx-1].
         */
        bool is_already_counted = false;
        if (nx > 0) { /* left */
            if (((nx-1) != x) || (ny != y)) {
                const string_id_t id = GET_STRING_ID(lst, nx-1, ny);
                if ((my_id == id) && (idx > 0)) {
                    for (int i = 0; i < idx; i++) {
                        if ((str->belong_id[i].x == (nx-1))
                            && (str->belong_id[i].y == ny)) {
                            is_already_counted = true;
                            break;
                        }
                    }
                }
            }
        }
        if (nx < (BOARD_SIZE - 1)) { /* right */
            if (((nx+1) != x) || (ny != y)) {
                const string_id_t id = GET_STRING_ID(lst, nx+1, ny);
                if ((my_id == id) && (idx > 0)) {
                    for (int i = 0; i < idx; i++) {
                        if ((str->belong_id[i].x == (nx+1))
                            && (str->belong_id[i].y == ny)) {
                            is_already_counted = true;
                            break;
                        }
                    }
                }
            }
        }
        if (ny > 0) { /* down */
            if ((nx != x) || ((ny-1) != y)) {
                const string_id_t id = GET_STRING_ID(lst, nx, ny-1);
                if ((my_id == id) && (idx > 0)) {
                    for (int i = 0; i < idx; i++) {
                        if ((str->belong_id[i].x == nx)
                            && (str->belong_id[i].y == (ny-1))) {
                            is_already_counted = true;
                            break;
                        }
                    }
                }
            }
        }
        if (ny < (BOARD_SIZE - 1)) { /* up */
            if ((nx != x) || ((ny+1) != y)) {
                const string_id_t id = GET_STRING_ID(lst, nx, ny+1);
                if ((my_id == id) && (idx > 0)) {
                    for (int i = 0; i < idx; i++) {
                        if ((str->belong_id[i].x == nx)
                            && (str->belong_id[i].y == (ny+1))) {
                            is_already_counted = true;
                            break;
                        }
                    }
                }
            }
        }
        if (!is_already_counted) {
            str->num_liberty++;
        }
    }

    return 0;
}


/**
 * This function destroys the string specified by \a id.
 *
 * On success, the string is destroyed. This function returns 0.
 * The status of its neighbor strings is also updated.
 * Otherwise this function returns a negative value.
 *
 * [What is updated after destruction?]
 * (1) All the neighbor strings of the given string
 *   - num_liberty is increased. (Need calculation.)
 *   - num_adjacent is decreased by one. (\a id is removed.)
 *   - adjacent_id[] becomes smaller. (\a id is removed.)
 *
 * \param lst  String list
 * \param id   String ID
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_string_list_destroy_string(struct go_string_list* lst, string_id_t id)
{
    struct go_string* str = GET_STRING_PTR(lst, id);

    /*
     * Update adjacent_id[] of neighbor strings
     */
    const int NUM_ADJ = str->num_adjacent;
    for (int i = 0; i < NUM_ADJ; i++) {
        const string_id_t nid = str->adjacent_id[i];
        struct go_string* nstr = GET_STRING_PTR(lst, nid);

        const int NUM_ADJ_NEXT = nstr->num_adjacent;
        for (int j = 0; j < NUM_ADJ_NEXT; j++) {
            if (nstr->adjacent_id[j] == id) {
                for (int k = j; k < NUM_ADJ_NEXT-1; k++) {
                    nstr->adjacent_id[k] = nstr->adjacent_id[k+1];
                }
                nstr->num_adjacent--;
                break;
            }
        }
    }

    /*
     * Update num_liberty of neighbor strings
     */
    const int BOARD_SIZE = GET_BOARD_SIZE(lst);
    const int SIZE = str->size;
    const string_id_t my_id = str->string_id;
    for (int i = 0; i < SIZE; i++) {
        string_id_t checked[4];
        const coord_t x = str->belong_id[i].x;
        const coord_t y = str->belong_id[i].y;

        int idx = 0;
        if (x > 0) {
            const string_id_t next_id = GET_STRING_ID(lst, x-1, y);
            if ((next_id != STRING_ID_INVALID) && (next_id != my_id)) {
                struct go_string* nstr = GET_STRING_PTR(lst, next_id);
                nstr->num_liberty++;
                checked[idx++] = next_id;
            }
        }
        if (x < (BOARD_SIZE - 1)) {
            const string_id_t next_id = GET_STRING_ID(lst, x+1, y);
            if ((next_id != STRING_ID_INVALID) && (next_id != my_id)) {
                bool is_checked = false;
                for (int j = 0; j < idx; j++) {
                    if (next_id == checked[j]) {
                        is_checked = true;
                        break;
                    }
                }
                if (!is_checked) {
                    struct go_string* nstr = GET_STRING_PTR(lst, next_id);
                    nstr->num_liberty++;
                    checked[idx++] = next_id;
                }
            }
        }
        if (y > 0) {
            const string_id_t next_id = GET_STRING_ID(lst, x, y-1);
            if ((next_id != STRING_ID_INVALID) && (next_id != my_id)) {
                bool is_checked = false;
                for (int j = 0; j < idx; j++) {
                    if (next_id == checked[j]) {
                        is_checked = true;
                        break;
                    }
                }
                if (!is_checked) {
                    struct go_string* nstr = GET_STRING_PTR(lst, next_id);
                    nstr->num_liberty++;
                    checked[idx++] = next_id;
                }
            }
        }
        if (y < (BOARD_SIZE - 1)) {
            const string_id_t next_id = GET_STRING_ID(lst, x, y+1);
            if ((next_id != STRING_ID_INVALID) && (next_id != my_id)) {
                bool is_checked = false;
                for (int j = 0; j < idx; j++) {
                    if (next_id == checked[j]) {
                        is_checked = true;
                        break;
                    }
                }
                if (!is_checked) {
                    struct go_string* nstr = GET_STRING_PTR(lst, next_id);
                    nstr->num_liberty++;
                    checked[idx++] = next_id;
                }
            }
        }
    }

    /*
     * Destroy the given string
     */
    str->string_id = STRING_ID_INVALID;
    for (int i = 0; i < SIZE; i++) {
        const coord_t x = str->belong_id[i].x;
        const coord_t y = str->belong_id[i].y;
        SET_STRING_ID(lst, x, y, STRING_ID_INVALID);
    }
    go_string_list_release_id(lst, id);

    return 0;
}


/**
 * This function returns the color of the specified string.
 *
 * On success, it returns 0, and stores the color to \a color.
 * Otherwise it returns a negative value.
 *
 * \param lst          String list
 * \param id           String ID
 * \param color        (OUT) Color of the given string
 * \param size         (OUT) Size of the given string
 * \param num_liberty  (OUT) Number of liberty points of the given string
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_string_list_get_info2(struct go_string_list* lst, string_id_t id,
                             stone_t* color, int* size, int* num_liberty)
{
    struct go_string* str = GET_STRING_PTR(lst, id);
    *color = str->color;
    *size = str->size;
    *num_liberty = str->num_liberty;

    return 0;
}


int go_string_list_allocate_id(struct go_string_list* lst)
{
    // XXX: tentative
    for (int i = 0; i < MAX_STRING_NUM; i++) {
        if (lst->id_bitmap_array[i] == false) {
            lst->id_bitmap_array[i] = true;
            return i;
        }
    }
    //printf("%s: ID is full (panic)\n", __FUNCTION__);
    return STRING_ID_INVALID; /* Cannot reach here */
}


int go_string_list_release_id(struct go_string_list* lst, int id)
{
    // XXX: tentative
    assert((0 <= id) && (id < MAX_STRING_NUM));
    assert(lst->id_bitmap_array[id] == true);

    lst->id_bitmap_array[id] = false;

    return 0;
}


/***
 *** For debugging
 ***/

/**
 * This function displays all the identifiers of the string. (for debugging)
 *
 * \retval 0   Success
 */
int go_string_list_display(struct go_string_list* lst)
{	
    const int BOARD_SIZE = lst->board_size;
    const int NUM_STRING = lst->num_string;

    {
        printf("    ");
        for (coord_t x = 0; x < BOARD_SIZE; x++) {
            if (x < 8) {
                printf("%3c", x + 'A'); /* A, B, ... H */
            } else {
                printf("%3c", x + 'A' + 1); /* J, K, ... */
            }
        }
        printf("\n");
    }

    for (coord_t y = 0; y < BOARD_SIZE; y++) {
        printf("[%2d] ", BOARD_SIZE - y);
        for (coord_t x = 0; x < BOARD_SIZE; x++) {
            if (go_string_list_get_id(lst, x, y) == STRING_ID_INVALID) {
                printf(" + ");
            } else {
                printf("%2d ", go_string_list_get_id(lst, x, y));
            }
        }
        printf("\n");
    }
    printf("\n");

    printf("String information (Num %d)\n", NUM_STRING);
    for (int idx = 0; idx < MAX_STRING_NUM; idx++) {
        if (lst->string_list[idx].string_id == STRING_ID_INVALID) {
            continue;
        }

        printf("----------------------------------------\n");
        printf("ID             %d\n", lst->string_list[idx].string_id);
        printf("Color          %d\n", lst->string_list[idx].color);
        printf("Size           %d\n", lst->string_list[idx].size);
        const int SIZE = lst->string_list[idx].size;
        printf("Belong List: ");
        for (int i = 0; i < SIZE; i++) {
            printf("(%d,%d) ", lst->string_list[idx].belong_id[i].x,
                   lst->string_list[idx].belong_id[i].y);
        }
        printf("\n");

        printf("Num Liberty    %d\n", lst->string_list[idx].num_liberty);
        printf("Num Adjacent   %d\n", lst->string_list[idx].num_adjacent);

        const int NUM_ADJACENT = lst->string_list[idx].num_adjacent;
        if (NUM_ADJACENT > 0) {
            printf("Adjacent List: ");
            for (int i = 0; i < NUM_ADJACENT; i++) {
                printf("%d ", lst->string_list[idx].adjacent_id[i]);
            }
            printf("\n");
        }
    }

    return 0;
}


/***
 *** Deprecated functions
 ***/

/**
 * This function clears all the data structures of strings.
 *
 * \retval 0   Success
 */
int go_string_list_clear(struct go_string_list* lst)
{
    const int BOARD_SIZE = lst->board_size;
	
    lst->num_string = 0;

    // XXX: tentative
    for (int i = 0; i < MAX_STRING_NUM; i++) {
        lst->id_bitmap_array[i] = false;
    }

    for (coord_t y = 0; y < BOARD_SIZE; y++) {
        for (coord_t x = 0; x < BOARD_SIZE; x++) {
            lst->board_array[y][x] = STRING_ID_INVALID;
        }
    }

    go_string_list_initialize_string_list(lst);

    return 0;
}


/**
 * This function gets the position of stones of the specified string.
 * The string is specified by \a color, \a x and \a y.
 *
 * On success, it returns 0. It also stores the stone positions to \a array
 * and the numbers of stones to \a num.
 * Otherwise it returns a negative value.
 *
 * \param lst
 * \param color      Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x          Stone position (X corrdinate)
 * \param y          Stone position (Y corrdinate)
 * \param array_num  Size of array
 * \param array      (OUT) Array of stone positions of the string
 * \param num        (OUT) Number of stones which belongs to the string
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_string_list_get_string_stone(struct go_string_list* lst, stone_t color,
                                    coord_t x, coord_t y, int array_num,
                                    coord2d_t* array, int* num)
{
    const int BOARD_SIZE = lst->board_size;

    assert((0 <= x) && (x < BOARD_SIZE) && (0 <= y) && (y < BOARD_SIZE));

    const string_id_t id = go_string_list_get_id(lst, x, y);
    if (id == STRING_ID_INVALID) {
        return -1;
    }

    stone_t str_color;
    int str_size;
    int str_liberty;
    go_string_list_get_info2(lst, id, &str_color, &str_size, &str_liberty);

    if (str_color != color) {
        return -1;
    }
    if (array_num < str_size) {
        return -4;
    }

    int count = 0;
    for (coord_t py = 0; py < BOARD_SIZE; py++) {
        for (coord_t px = 0; px < BOARD_SIZE; px++) {
            if (lst->board_array[py][px] == id) {
                array[count].x = px;
                array[count].y = py;
                count++;
            }
        }
    }
    assert(count == str_size);

    *num = str_size;

    return 0;
}


/**
 * This function returns the number of liberty points of the specified string.
 *
 * On success, it returns 0, and stores the number of libery points to \a num.
 * Otherwise it returns a negative value.
 *
 * \param lst
 * \param color  Stone color (STONE_COLOR_BLACK or STONE_COLOR_WHITE)
 * \param x      Stone position (X corrdinate)
 * \param y      Stone position (Y corrdinate)
 * \param num    (OUT) Number of liberty points
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int go_string_list_get_liberty_number(struct go_string_list* lst,
                                      stone_t color, coord_t x, coord_t y,
                                      int* num)
{
    const int BOARD_SIZE = lst->board_size;

    assert((0 <= x) && (x < BOARD_SIZE) && (0 <= y) && (y < BOARD_SIZE));

    const string_id_t id = go_string_list_get_id(lst, x, y);
    if (id == STRING_ID_INVALID) {
        return -2;
    }

    stone_t str_color;
    int str_size;
    int str_liberty;
    go_string_list_get_info2(lst, id, &str_color, &str_size, &str_liberty);

    if (str_color != color) {
        return -1;
    }

    *num = str_liberty;

    return 0;
}


/**
 * This function gets the information on the specified string.
 *
 * \param id      String ID
 * \param string  (OUT) Information on the given string
 *
 * \retval 0   Success
 * \retval -1  \a id is invalid.
 */
int go_string_list_get_info(struct go_string_list* lst, string_id_t id,
                            struct go_string* string)
{
    if ((id < 0) || (MAX_STRING_NUM <= id)) {
        return -1;
    } else {
        assert((0 <= id) && (id < MAX_STRING_NUM));
        *string = lst->string_list[id];
        return 0;
    }
}


/**
 * This function returns the identifier of the string which is
 * specified by \a x and \a y.
 *
 * \retval valid_id           Identifier of the string
 * \retval STRING_ID_INVALID  An empty position is specified.
 */
string_id_t go_string_list_get_id(struct go_string_list* lst, coord_t x,
                                  coord_t y)
{
    return lst->board_array[y][x];
}


/***
 *** Internal functions
 ***/

/** \internal
 * This function initializes the string list.
 *
 * \retval 0   Success
 */
int go_string_list_initialize_string_list(struct go_string_list* lst)
{
    for (int idx = 0; idx < MAX_STRING_NUM; idx++) {
        lst->string_list[idx].string_id = STRING_ID_INVALID;
        lst->string_list[idx].color = STONE_COLOR_EMPTY;
        lst->string_list[idx].size = 0;
        lst->string_list[idx].num_liberty = 0;
        lst->string_list[idx].num_adjacent = 0;

#ifdef DIAGNOSTIC
        for (int i = 0; i < MAX_STRING_SIZE; i++) {
            lst->string_list[idx].belong_id[i].x = STONE_COORD_INVALID;
            lst->string_list[idx].belong_id[i].y = STONE_COORD_INVALID;
        }
        for (int i = 0; i < MAX_ADJACENT_NUM; i++) {
            lst->string_list[idx].adjacent_id[i] = STRING_ID_INVALID;
        }
#endif /* DIAGNOSTIC */
    }

    return 0;
}
