/*
 * GTP commands (Version 2) are implemented in this file.
 *
 * The supported GTP commands are the followings:
 *
 *     01. protocol_version
 *     02. name
 *     03. version
 *     04. known_command
 *     05. list_commands
 *     06. quit
 *     07. boardsize
 *     08. clear_board
 *     09. komi
 *     10. play
 *     11. genmove
 *     12. time_settings
 *     13. loadsgf
 */

#include <assert.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "go_engine.h"
#include "go_types.h"
#include "gtp_command.h"
#include "utility.h"


#define GTP_COMMAND_PROTOCOL_VERSION    "2"
#define GTP_COMMAND_ENGINE_NAME         "yayoi"
#define GTP_COMMAND_ENGINE_VERSION      "0.90"

#define GTP_COMMAND_ID_INVALID          -1

#define GTP_COMMAND_COLOR_LENGTH        5  /* "BLACK" or "WHITE" */
#define GTP_COMMAND_COLOR_LENGTH_SHORT  1  /* "B" or "W" */

#define GTP_COMMAND_BOARD_SIZE_MINIMAL  5
#define GTP_COMMAND_BOARD_SIZE_SMALL    9
#define GTP_COMMAND_BOARD_SIZE_MEDIUM   13
#define GTP_COMMAND_BOARD_SIZE_LARGE    19

#define BUFSIZE                         256

#define NUM_GTP_COMMAND                 13
#define MAX_GTP_COMMAND_LENGTH          32

#define MAX_VERTEX_LENGTH               8


typedef int (*gtp_func_t)(int, char*);

typedef struct gtp_command {
    const char name[MAX_GTP_COMMAND_LENGTH];
    gtp_func_t function;
} gtp_command_t;


/*
 * Command functions
 */
static int gtp_command_get_protocol_version(int, char*);
static int gtp_command_get_engine_name(int, char*);
static int gtp_command_get_engine_version(int, char*);
static int gtp_command_is_known_command(int, char*);
static int gtp_command_get_command_list(int, char*);
static int gtp_command_quit(int, char*);
static int gtp_command_set_boardsize(int, char*);
static int gtp_command_clear_board(int, char*);
static int gtp_command_set_komi(int, char*);
static int gtp_command_set_opponent_move(int, char*);
static int gtp_command_get_next_move(int, char*);
static int gtp_command_set_time_settings(int, char*);
static int gtp_command_load_sgf_file(int, char*);


/*
 * Sub-functions
 */
static int gtp_command_success(int, const char*, ...);
static int gtp_command_failure(int, const char*, ...);
static int gtp_command_encode_move(move_t, coord_t, coord_t, char*);
static int gtp_command_decode_move(char*, move_t*, coord_t*, coord_t*);


/*
 * Command list
 */
static gtp_command_t g_gtp_command[NUM_GTP_COMMAND] = {
    {"protocol_version",        gtp_command_get_protocol_version},
    {"name",                    gtp_command_get_engine_name},
    {"version",                 gtp_command_get_engine_version},
    {"known_command",           gtp_command_is_known_command},
    {"list_commands",           gtp_command_get_command_list},
    {"quit",                    gtp_command_quit},
    {"boardsize",               gtp_command_set_boardsize},
    {"clear_board",             gtp_command_clear_board},
    {"komi",                    gtp_command_set_komi},
    {"play",                    gtp_command_set_opponent_move},
    {"genmove",                 gtp_command_get_next_move},
    {"time_settings",           gtp_command_set_time_settings},
    {"loadsgf",                 gtp_command_load_sgf_file},
};


/*
 * main loop
 */
int gtp_command_main(void)
{
    char line[BUFSIZE];
    char command[BUFSIZE];

    for (;;) {
        if (!fgets(line, BUFSIZE, stdin)) {
            break;
        }

        /*
         * Cut '\n' from line[]
         */
        char* p = strchr(line, '\n');
        if (p != NULL) {
            *p = '\0';
        }
        p = line;

        /*
         * Get identifier
         */
        int id, num;
        if (sscanf(p, "%d%n", &id, &num) == 1) {
            p += num;
        } else {
            id = GTP_COMMAND_ID_INVALID;
        }

        /*
         * Get GTP command's name
         */
        if (sscanf(p, " %s %n", command, &num) < 1) {
            continue;
        }
        p += num;

        /*
         * Apply GTP command
         */
        int status = 0;
        for (int i = 0; i < NUM_GTP_COMMAND; i++) {
            if (strncmp(command, g_gtp_command[i].name,
                        MAX_GTP_COMMAND_LENGTH) == 0) {
                status = (*g_gtp_command[i].function)(id, p);
                break;
            }
            if (i == (NUM_GTP_COMMAND - 1)) {
                char errmsg[BUFSIZE];
                snprintf(errmsg, BUFSIZE, "Unknown command: %s", command);
                gtp_command_failure(id, errmsg);
            }
        }

        if (status != 0) {
            break;
        }
    }

    return 0;
}


/***
 *** GTP commands
 ***/

/*
 * protocol_version
 */
int gtp_command_get_protocol_version(int id, char* param)
{
    (void)param;

    return gtp_command_success(id, GTP_COMMAND_PROTOCOL_VERSION);
}


/*
 * name
 */
int gtp_command_get_engine_name(int id, char* param)
{
    (void)param;

    return gtp_command_success(id, GTP_COMMAND_ENGINE_NAME);
}


/*
 * version
 */
int gtp_command_get_engine_version(int id, char* param)
{
    (void)param;

    return gtp_command_success(id, GTP_COMMAND_ENGINE_VERSION);
}


/*
 * known_command command
 */
int gtp_command_is_known_command(int id, char* param)
{
    for (int i = 0; i < NUM_GTP_COMMAND; i++) {
        if (strcmp(param, g_gtp_command[i].name) == 0) {
            return gtp_command_success(id, "true");
        }
    }

    return gtp_command_success(id, "false");
}


/*
 * list_commands
 */
int gtp_command_get_command_list(int id, char* param)
{
    char msg[NUM_GTP_COMMAND * MAX_GTP_COMMAND_LENGTH]; // XXX

    (void)param;

    char* p = msg;
    for (int i = 0; i < NUM_GTP_COMMAND; i++) {
        snprintf(p, MAX_GTP_COMMAND_LENGTH, "%s", g_gtp_command[i].name);
        p += strlen(g_gtp_command[i].name);

        /*
         * All commands except for the last one need line-feeds.
         * The last one has to avoid it because the GTP protocol does not
         * permit successive triple line-feeds.
         */
        if (i < (NUM_GTP_COMMAND - 1)) {
            *(p++) = '\n';
        }
    }
    *p = '\0';

    return gtp_command_success(id, (const char*)msg);
}


/*
 * quit
 */
int gtp_command_quit(int id, char* param)
{
    (void)param;

    gtp_command_success(id, "");
    return -1;
}


/*
 * boardsize size
 */
int gtp_command_set_boardsize(int id, char* param)
{
    int rv;
    int boardsize;

    rv = sscanf(param, "%d", &boardsize);
    if (rv < 1) {
        return gtp_command_failure(id, "Bad board size");
    }

    /*
     * Only 5x5, 9x9, 13x13 or 19x19 are supported.
     */
    if ((boardsize != GTP_COMMAND_BOARD_SIZE_MINIMAL) &&
        (boardsize != GTP_COMMAND_BOARD_SIZE_SMALL) &&
        (boardsize != GTP_COMMAND_BOARD_SIZE_MEDIUM) &&
        (boardsize != GTP_COMMAND_BOARD_SIZE_LARGE)) {
        return gtp_command_failure(id, "Unsupported board size");
    }

    /*
     * Pass board size to the engine.
     */
    rv = go_engine_set_boardsize(boardsize);
    if (rv != 0) {
        return gtp_command_failure(id, "Cannot set board size");
    }

    return gtp_command_success(id, "");
}


/*
 * clear_board
 */
int gtp_command_clear_board(int id, char* param)
{
    int rv;

    (void)param;

    /*
     * Ask the engine to clear board.
     */
    rv = go_engine_reinitialize();
    if (rv != 0) {
        return gtp_command_failure(id, "Cannot clear board");
    } else {
        return gtp_command_success(id, "");
    }
}


/*
 * komi fval
 */
int gtp_command_set_komi(int id, char* param)
{
    int rv;
    float komi;

    rv = sscanf(param, "%f", &komi);
    if (rv < 1) {
        return gtp_command_failure(id, "Bad komi value");
    }

    /*
     * Pass komi to the engine.
     */
    rv = go_engine_set_komi(komi);
    if (rv != 0) {
        return gtp_command_failure(id, "Cannot set komi");
    }

    return gtp_command_success(id, "");
}


/*
 * play color vertex
 */
int gtp_command_set_opponent_move(int id, char* param)
{
    int rv;
    move_t flags;
    coord_t x, y;
    char* p;

    /*
     * Cut color field. It may be short or normal.
     */
    const size_t SIZE_SHORT = GTP_COMMAND_COLOR_LENGTH_SHORT + 1;
    const size_t SIZE_NORMAL = GTP_COMMAND_COLOR_LENGTH + 1;
    if ((strncmp(param, "B ", SIZE_SHORT) == 0) ||
        (strncmp(param, "W ", SIZE_SHORT) == 0) ||
        (strncmp(param, "b ", SIZE_SHORT) == 0) ||
        (strncmp(param, "w ", SIZE_SHORT) == 0)) {
        p = param + SIZE_SHORT;
    } else if ((strncmp(param, "BLACK ", SIZE_NORMAL) == 0) ||
               (strncmp(param, "WHITE ", SIZE_NORMAL) == 0) ||
               (strncmp(param, "black ", SIZE_NORMAL) == 0) ||
               (strncmp(param, "white ", SIZE_NORMAL) == 0)) {
        p = param + SIZE_NORMAL;
    } else {
        return gtp_command_failure(id, "Invalid color");
    }

    /*
     * Get move info by parsing param.
     */
    rv = gtp_command_decode_move(p, &flags, &x, &y);
    if (rv != 0) {
        return gtp_command_failure(id, "Invalid move");
    }

    /*
     * Pass move info to the engine.
     */
    go_engine_set_opponent_move(flags, x, y);

    return gtp_command_success(id, "");
}


/*
 * genmove color
 */
int gtp_command_get_next_move(int id, char* param)
{
    move_t flags;
    coord_t x, y;
    char msg[MAX_VERTEX_LENGTH];

    (void)param;

#ifdef notyet
    /*
     * Check the stone color of the engine.
     */
#endif /* notyet */

    go_engine_get_next_move(&flags, &x, &y);

    gtp_command_encode_move(flags, x, y, msg);

    return gtp_command_success(id, (const char*)msg);
}


/*
 * time_settings main_time byo_yomi_time byo_yomi_stones
 */
int gtp_command_set_time_settings(int id, char* param)
{
    int rv;
    int main_time;
    int additional_time;
    int num_stone;

    rv = sscanf(param, "%d %d %d", &main_time, &additional_time, &num_stone);
    if (rv < 1) {
        return gtp_command_failure(id, "Bad time-settings");
    }

    if (num_stone != 1) {
        return gtp_command_failure(id, "Bad byo-yomi stones");
    }
    if (additional_time <= 0) {
        return gtp_command_failure(id, "Bad byo-yomi time");
    }
    if (main_time != 0) {
        return gtp_command_failure(id, "Main time not supported");
    }

    /*
     * Specify time settings to the engine.
     */
    rv = go_engine_set_time_settings(0, additional_time * 1000 * 1000);
    if (rv != 0) {
        return gtp_command_failure(id, "Cannot set time-settings");
    }

    return gtp_command_success(id, "");
}


/*
 * loadsgf filename move_number
 */
int gtp_command_load_sgf_file(int id, char* param)
{
    int rv;
    const char* filename = param;

#ifdef notyet
    /*
     * XXX: move_number is ignored at this point.
     */
#endif /* notyet */

    /*
     * Load the specified SGF file.
     */
    rv = go_engine_load_sgf_file(filename, MOVE_COUNT_INVALID);
    if (rv != 0) {
        return gtp_command_success(id, "cannot load file");
    }

    return gtp_command_success(id, "");
}


/***
 *** Sub-functions
 ***/

int gtp_command_success(int id, const char* format, ...)
{
    printf("=");
    if (id < 0) {
        printf(" ");
    } else {
        printf("%d ", id);
    }
    printf("%s", format);
    printf("\n\n");
    fflush(NULL);

    return 0;
}


int gtp_command_failure(int id, const char* format, ...)
{
    printf("?");
    if (id < 0) {
        printf(" ");
    } else {
        printf("%d ", id);
    }
    printf("%s", format);
    printf("\n\n");
    fflush(NULL);

    return 0;
}


/** \internal
 * This function creates a string for the given move.
 *
 * \param flags  flags of the move (MOVE_FLAGS_{PASS,RESIGN,NORMAL})
 * \param x      x-coordinate of the move
 * \param y      y-coordinate of the move
 * \param mvstr  (OUT) string for the move
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int gtp_command_encode_move(move_t flags, coord_t x, coord_t y, char* mvstr)
{
#ifdef notyet
    /*
     * Sanity check
     */
#endif /* notyet */

    if (flags == MOVE_FLAGS_PASS) {
        snprintf(mvstr, MAX_VERTEX_LENGTH, "PASS");
    } else if (flags == MOVE_FLAGS_RESIGN) {
        snprintf(mvstr, MAX_VERTEX_LENGTH, "RESIGN");
    } else {
        assert(flags == MOVE_FLAGS_NORMAL);

        if (x < 8) { /* A, B, C, D, E, F, G, H */
            snprintf(mvstr, MAX_VERTEX_LENGTH, "%c%d", 'A'+x, y+1);
        } else { /* J, K, L, ... */
            snprintf(mvstr, MAX_VERTEX_LENGTH, "%c%d", 'A'+x+1, y+1);
        }
    }

    return 0;
}


/** \internal
 * This function decodes the given string and gets the move info.
 *
 * Either of "pass" ("PASS"), "resign" ("RESIGN") or a pair of coodinates
 * is expected as the given string.
 *
 * \param mvstr  string for the move
 * \param flags  (OUT) flags of the move (MOVE_FLAGS_{PASS,RESIGN,NORMAL})
 * \param x      (OUT) x-coordinate of the move
 * \param y      (OUT) y-coordinate of the move
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int gtp_command_decode_move(char* mvstr, move_t* flags, coord_t* x, coord_t* y)
{
    if ((strncmp(mvstr, "pass", MAX_VERTEX_LENGTH) == 0) ||
        (strncmp(mvstr, "PASS", MAX_VERTEX_LENGTH) == 0)) {
        *flags = MOVE_FLAGS_PASS;
        *x = STONE_COORD_INVALID;
        *y = STONE_COORD_INVALID;
    } else if ((strncmp(mvstr, "resign", MAX_VERTEX_LENGTH) == 0) ||
               (strncmp(mvstr, "RESIGN", MAX_VERTEX_LENGTH) == 0)) {
        *flags = MOVE_FLAGS_RESIGN;
        *x = STONE_COORD_INVALID;
        *y = STONE_COORD_INVALID;
    } else {
        int rv;
        char xc;
        coord_t yc;

        rv = sscanf(mvstr, "%c%d", &xc, &yc);
        if (rv < 1) {
            return -1;
        }

        *flags = MOVE_FLAGS_NORMAL;
        if (toupper(xc) < 'I') {
            *x = toupper(xc) - 'A';
        } else {
            *x = toupper(xc) - 'A' - 1;
        }
        *y = yc - 1;
    }

    return 0;
}
