#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "go_engine.h" // go_engine_reinitialize()
#include "go_engine_internal.h"
#include "sgf_loader.h"


#define SGF_LOADER_BUFSIZE      4096


/*
 * Internal functions
 */
static int sgf_file_get_board_size(char*, int, int*);
static int sgf_file_get_komi(char*, int, float*);
static int sgf_file_get_black_stone_list(char*, int, coord2d_t*, int, int*);
static int sgf_file_get_white_stone_list(char*, int, coord2d_t*, int, int*);


/**
 * This function loads the specified SGF file.
 * It just parses some limited fields and updates the board status.
 *
 * \param filename  SGF filename
 * \param move_num  Number of moves
 *
 * \retval 0   Success
 * \retval -1  Failure
 *
 * \note This function parses SZ (size), KM (komi), AB (add black) and
 *       AW (add white) only.
 */
int sgf_file_load_simple(const char* filename, int* board_size, float* komi,
                         coord2d_t* blist, int* blist_num, coord2d_t* wlist,
                         int* wlist_num)
{
    int rv;
    char buf[SGF_LOADER_BUFSIZE];

    FILE* fp = fopen(filename, "r");
    if (fp == NULL) {
        return -1;
    }

    const size_t size = fread(buf, SGF_LOADER_BUFSIZE, 1, fp);
    assert(size < SGF_LOADER_BUFSIZE);
    /* If size == BUFSIZE, buf[] may be too small. */

    /*
     * Get board size (SZ)
     */
    {
        int sz;
        rv = sgf_file_get_board_size(buf, size, &sz);
        if (rv != 0) {
            return -1;
        }
        *board_size = sz;
    }

    /*
     * Get komi (KM)
     */
    {
        float km;
        rv = sgf_file_get_komi(buf, size, &km);
        if (rv != 0) {
            return -1;
        }
        *komi = km;
    }

    /*
     * Get stone positions (AB; black stones)
     */
    {
        const int blist_max_num = *blist_num;
        rv = sgf_file_get_black_stone_list(buf, size, blist, blist_max_num,
                                           blist_num);
        if (rv != 0) {
            return -1;
        }
    }

    /*
     * Get stone positions (AW; white stones)
     */
    {
        const int wlist_max_num = *wlist_num;
        rv = sgf_file_get_white_stone_list(buf, size, wlist, wlist_max_num,
                                           wlist_num);
        if (rv != 0) {
            return -1;
        }
    }

    fclose(fp);

    return 0;
}


/** \internal
 * This function parses the given string and gets the board size. (SZ field)
 *
 * \param buf         String of SGF file
 * \param buf_size    Size of string
 * \param board_size  (OUT) Board size
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int sgf_file_get_board_size(char* buf, int buf_size, int* board_size)
{
    (void)buf_size;

    for (char* p = buf; *p != '\0'; p++) {
        if (strncmp(p, "SZ", 2) == 0) {
            char str[3];
            memcpy(str, p+3, 2);
            str[2] = '\0';

            for (int i = 0; i < 3; i++) {
                if (str[i] == ']') {
                    str[i] = '\0';
                }
            }

            *board_size = atoi(str);
            return 0;
        }
    }

    return -1;
}


/** \internal
 * This function parses the given string and gets the value of komi. (KM field)
 *
 * \param buf       String of SGF file
 * \param buf_size  Size of string
 * \param komi      (OUT) Value of komi
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int sgf_file_get_komi(char* buf, int buf_size, float* komi)
{
    (void)buf_size;

    for (char* p = buf; *p != '\0'; p++) {
        if (strncmp(p, "KM", 2) == 0) {
            char str[10]; // XXX: assume length of value
            memcpy(str, p+3, 9);
            str[9] = '\0';

            for (int i = 0; i < 10; i++) {
                if (str[i] == ']') {
                    str[i] = '\0';
                }
            }

            *komi = (float)atof(str);
            return 0;
        }
    }

    return -1;
}


/** \internal
 * This function parses the given string and gets the list of positions of
 * black stones. (AB field)
 *
 * \param buf         String of SGF file
 * \param buf_size    Size of string
 * \param blist       (OUT) Array of position of black stone
 * \param blist_num   Size of blist[]
 * \param actual_num  (OUT) Actual number of stones
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int sgf_file_get_black_stone_list(char* buf, int buf_size, coord2d_t* blist,
                                  int blist_num, int* actual_num)
{
    int num;

    (void)buf_size;

    for (char* p = buf; *p != '\0'; p++) {
        if (strncmp(p, "AB", 2) == 0) {
            p += 2;

            num = 0;
            for (int i = 0; i < blist_num; i++) {
                if (*p != '[') {
                    *actual_num = num;
                    return 0;
                }
                const char xc = *(p+1);
                const char yc = *(p+2);
                if (*(p+3) != ']') {
                    return -1;
                }

                blist[i].x = xc - 'a';
                blist[i].y = yc - 'a';
                num++;
                p += 4;
            }
        }
    }

    return -1;
}


/** \internal
 * This function parses the given string and gets the list of positions of
 * white stones. (AW field)
 *
 * \param buf         String of SGF file
 * \param buf_size    Size of string
 * \param wlist       (OUT) Array of position of white stone
 * \param wlist_num   Size of blist[]
 * \param actual_num  (OUT) Actual number of stones
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int sgf_file_get_white_stone_list(char* buf, int buf_size, coord2d_t* wlist,
                                  int wlist_num, int* actual_num)
{
    int num;

    (void)buf_size;

    for (char* p = buf; *p != '\0'; p++) {
        if (strncmp(p, "AW", 2) == 0) {
            p += 2;

            num = 0;
            for (int i = 0; i < wlist_num; i++) {
                if (*p != '[') {
                    *actual_num = num;
                    return 0;
                }

                const char xc = *(p+1);
                const char yc = *(p+2);
                if (*(p+3) != ']') {
                    return -1;
                }

                wlist[i].x = xc - 'a';
                wlist[i].y = yc - 'a';
                num++;
                p += 4;
            }
        }
    }

    return -1;
}
