#include "go_types.h"
#include "uct_tree.h"

#include <assert.h>
#include <math.h>
#ifndef __cplusplus
#include <stdbool.h>
#endif /* __cplusplus */
#include <stdio.h>


#define NODE_INITIALIZE(nd)                     \
    do {                                        \
        (nd)->value = 0;                        \
        (nd)->num_win = 0;                      \
        (nd)->num_match = 0;                    \
        (nd)->num_children = 0;                 \
        (nd)->x = STONE_COORD_INVALID;          \
        (nd)->y = STONE_COORD_INVALID;          \
        (nd)->depth = 0;                        \
        (nd)->parent = NULL;                    \
        (nd)->next = NULL;                      \
        (nd)->children = NULL;                  \
    } while (0);


/**
 * This function initialzes the given tree.
 *
 * \param tree  Tree
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int uct_tree_initialize(struct uct_tree* tree)
{
    int rv;

    tree->node_idx = 0;

    struct uct_node* root;
    rv = uct_tree_create_node(tree, &root);
    if (rv != 0) {
        return -1;
    }

    NODE_INITIALIZE(root);
    tree->root = root;

    return 0;
}


/**
 * This function terminates the given tree.
 *
 * \param tree  Tree
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int uct_tree_terminate(struct uct_tree* tree)
{
    tree->root = NULL;
    tree->node_idx = 0; /* delete all the nodes */

    return 0;
}


/**
 * This function creates a new node structure.
 *
 * \param tree   Tree
 * \param nnode  (OUT) New node structure
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int uct_tree_create_node(struct uct_tree* tree, struct uct_node** nnode)
{
    const int idx = tree->node_idx;

    assert(idx < UCT_NUM_NODE);

    if (idx >= UCT_NUM_NODE) {
        return -1;
    }

    struct uct_node* new_node = &(tree->node_array[idx]);
    tree->node_idx++;

    NODE_INITIALIZE(new_node);

    assert(new_node->value == 0);
    assert(new_node->num_win == 0);
    assert(new_node->num_match == 0);
    assert(new_node->num_children == 0);
    assert(new_node->x == STONE_COORD_INVALID);
    assert(new_node->y == STONE_COORD_INVALID);
    assert(new_node->parent == NULL);
    assert(new_node->next == NULL);
    assert(new_node->children == NULL);

    *nnode = new_node;

    return 0;
}


/**
 * This function adds \a cnode to \a tr as a child of \a pnode.
 *
 * \param tree   Tree
 * \param pnode  Parent node
 * \param cnode  Child node
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int uct_tree_add_child(struct uct_tree* tree, struct uct_node* pnode,
                       struct uct_node* cnode)
{
    (void)tree;

    pnode->num_children++;
    cnode->depth = pnode->depth + 1;

    if (pnode->children == NULL) {
        /*
         * This is the first child.
         */
        pnode->children = cnode;
    } else {
        /*
         * Append \a cnode to the children's list.
         */
        struct uct_node* child = pnode->children;
        while (1) {
            if (child->next == NULL) {
                break;
            }

            child = child->next;
        }
        child->next = cnode;
    }

    cnode->parent = pnode;

    return 0;
}


/**
 * This function searches the maximum child of the given node.
 *
 * On success, it returns 0. It also stores the maximum child to \a cnode.
 * Otherwise it returns a negative value.
 *
 * \param tree   Tree
 * \param pnode  Parent node
 * \param cnode  (OUT) Maximum child node
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int uct_tree_get_max_child(struct uct_tree* tree, struct uct_node* pnode,
                           struct uct_node** cnode)
{
    if (pnode->children == NULL) {
        return -1;
    }

    double max_value = -HUGE_VAL; /* XXX: Does child->value take [0,1]? */ 
    struct uct_node* max_child = NULL;
    for (struct uct_node* child = pnode->children; child != NULL;
         child = child->next) {
        if (child->value > max_value) {
            max_value = child->value;
            max_child = child;
        }
    }

    *cnode = max_child;

    return 0;
}


/**
 * This function searches the maximum child of the given node.
 *
 * Also the limitation of num_match is considered.
 * If num_match is less than \a limit, a child is ignored.
 *
 * If num_match of all the children is less than \a limit, this function
 * returns an error. If there is at least a child which satisfies the
 * limitation, this function successes.
 *
 * On success, it returns 0. It also stores the maximum child to \a cnode.
 * Otherwise it returns a negative value.
 *
 * \param tree   Tree
 * \param pnode  Parent node
 * \param cnode  (OUT) Maximum child node which satisfies the limitaiton
 *
 * \retval 0   Success
 * \retval -1  Failure
 * \retval -2  Failure
 */
int uct_tree_get_max_child_with_limit(struct uct_tree* tree,
                                      struct uct_node* pnode, int limit,
                                      struct uct_node** cnode)
{
    if (pnode->children == NULL) {
        return -1;
    }

    int num_found = 0;
    double max_value = -HUGE_VAL; /* XXX: Does child->value take [0,1]? */ 
    struct uct_node* max_child = NULL;
    for (struct uct_node* child = pnode->children; child != NULL;
         child = child->next) {
        if (child->num_match < limit) {
            continue;
        }
        if (child->value > max_value) {
            max_value = child->value;
            max_child = child;
            num_found++;
        }
    }

    if (num_found > 0) {
        *cnode = max_child;
        return 0;
    } else {
        /*
         * No children satisfies the limitation.
         */
        return -2;
    }
}


/**
 * This function searches the top 5 children of the given node.
 *
 * On success, it returns 0. It also stores the top 5 child to \a carray.
 * The actual number of elements of \a carray is stored to \a num_carray.
 * Otherwise it returns a negative value.
 *
 * \param tree        Tree
 * \param pnode       Parent node
 * \param carray      (OUT) Top 5 children (Array of node)
 * \param num_carray  (OUT) Number of actual elements of \a carray
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int uct_tree_get_top5_children(struct uct_tree* tree, struct uct_node* pnode,
                               struct uct_node** carray, int* num_carray)
{
    if (pnode->children == NULL) {
        return -1;
    }

    const int NUM_ARRAY = 5;
    double max_value[NUM_ARRAY];
    struct uct_node* max_child[NUM_ARRAY];
    for (int i = 0; i < NUM_ARRAY; i++) {
        max_value[i] = -HUGE_VAL;
        max_child[i] = NULL;
    }

    int idx = 0;
    for (struct uct_node* child = pnode->children; child != NULL;
         child = child->next) {
        bool is_updated = false;
        if (idx < NUM_ARRAY) {
            if (child->value > max_value[idx]) {
                max_value[idx] = child->value;
                max_child[idx] = child;
                is_updated = true;
                idx++;
            }
        } else {
            if (child->value > max_value[NUM_ARRAY-1]) {
                max_value[NUM_ARRAY-1] = child->value;
                max_child[NUM_ARRAY-1] = child;
                is_updated = true;
            }
        }

        /*
         * If a new child is inserted, sort max_value[] and max_child[]
         * in increasing order.
         */
        if (is_updated) {
            for (int i = 0; i < idx; i++) {
                for (int j = i+1; j < idx; j++) {
                    if (max_value[i] < max_value[j]) {
                        double tmp_value = max_value[j];
                        max_value[j] = max_value[i];
                        max_value[i] = tmp_value;

                        struct uct_node* tmp_child = max_child[j];
                        max_child[j] = max_child[i];
                        max_child[i] = tmp_child;
                    }
                }
            }
        }
    }

    for (int i = 0; i < idx; i++) {
        carray[i] = max_child[i];
    }
    *num_carray = idx;

    return 0;
}


/**
 * This function searches the top 5 children of the given node.
 *
 * Also the limitation of num_match is considered.
 * If num_match is less than \a limit, a child is ignored.
 *
 * If num_match of all the children is less than \a limit, this function
 * returns an error. If there is at least a child which satisfies the
 * limitation, this function successes.
 *
 * On success, it returns 0. It also stores the top 5 child to \a carray.
 * The actual number of elements of \a carray is stored to \a num_carray.
 * Otherwise it returns a negative value.
 *
 * \param tree        Tree
 * \param pnode       Parent node
 * \param limit       Limitation
 * \param carray      (OUT) Top 5 children (Array of node)
 * \param num_carray  (OUT) Number of actual elements of \a carray
 *
 * \retval 0   Success
 * \retval -1  Failure
 * \retval -2  Failure
 */
int uct_tree_get_top5_children_with_limit(struct uct_tree* tree,
                                          struct uct_node* pnode, int limit,
                                          struct uct_node** carray,
                                          int* num_carray)
{
    if (pnode->children == NULL) {
        return -1;
    }

    const int NUM_ARRAY = 5;
    double max_value[NUM_ARRAY];
    struct uct_node* max_child[NUM_ARRAY];
    for (int i = 0; i < NUM_ARRAY; i++) {
        max_value[i] = -HUGE_VAL;
        max_child[i] = NULL;
    }

    int idx = 0;
    for (struct uct_node* child = pnode->children; child != NULL;
         child = child->next) {
        if (child->num_match < limit) {
            continue;
        }

        bool is_updated = false;
        if (idx < NUM_ARRAY) {
            if (child->value > max_value[idx]) {
                max_value[idx] = child->value;
                max_child[idx] = child;
                is_updated = true;
                idx++;
            }
        } else {
            if (child->value > max_value[NUM_ARRAY-1]) {
                max_value[NUM_ARRAY-1] = child->value;
                max_child[NUM_ARRAY-1] = child;
                is_updated = true;
            }
        }

        /*
         * If a new child is inserted, sort max_value[] and max_child[]
         * in increasing order.
         */
        if (is_updated) {
            for (int i = 0; i < idx; i++) {
                for (int j = i+1; j < idx; j++) {
                    if (max_value[i] < max_value[j]) {
                        double tmp_value = max_value[j];
                        max_value[j] = max_value[i];
                        max_value[i] = tmp_value;

                        struct uct_node* tmp_child = max_child[j];
                        max_child[j] = max_child[i];
                        max_child[i] = tmp_child;
                    }
                }
            }
        }
    }

    if (idx == 0) {
        return -2;
    }

    /*
     * Store the results
     */
    for (int i = 0; i < idx; i++) {
        carray[i] = max_child[i];
    }
    *num_carray = idx;

    return 0;
}


/**
 * This function dumps the subtree whose root is \a node.
 *
 * \param tree  Tree
 * \param node  A node of \a tree. This acts as the root of the subtree.
 *
 * \retval 0   Success
 * \retval -1  Failure
 */
int uct_tree_dump_tree(struct uct_tree* tree, struct uct_node* node)
{
    printf("----------------------------------------\n");
    printf("Node (%d,%d) [Depth %d, Total %d]\n", node->x, node->y,
           node->depth, tree->node_idx);
    printf("    Win rate      %f\n", node->value);
    printf("    # of wins     %d\n", node->num_win);
    printf("    # of matches  %d\n", node->num_match);
    printf("    # of children %d\n", node->num_children);

    if (node->children != NULL) {
        for (struct uct_node* child = node->children; child != NULL;
             child = child->next) {
            uct_tree_dump_tree(tree, child);
        }
    }

    return 0;
}
